"""Deprecated image preprocessing APIs from Keras 1."""

import collections
import multiprocessing
import os
import threading
import warnings

import numpy as np

from keras.src import backend
from keras.src.api_export import keras_export
from keras.src.trainers.data_adapters.py_dataset_adapter import PyDataset
from keras.src.utils import image_utils
from keras.src.utils import io_utils
from keras.src.utils.module_utils import scipy


@keras_export("keras._legacy.preprocessing.image.Iterator")
class Iterator(PyDataset):
    """Base class for image data iterators.

    DEPRECATED.

    Every `Iterator` must implement the `_get_batches_of_transformed_samples`
    method.

    Args:
        n: Integer, total number of samples in the dataset to loop over.
        batch_size: Integer, size of a batch.
        shuffle: Boolean, whether to shuffle the data between epochs.
        seed: Random seeding for data shuffling.
    """

    white_list_formats = ("png", "jpg", "jpeg", "bmp", "ppm", "tif", "tiff")

    def __init__(self, n, batch_size, shuffle, seed):
        self.n = n
        self.batch_size = batch_size
        self.seed = seed
        self.shuffle = shuffle
        self.batch_index = 0
        self.total_batches_seen = 0
        self.lock = threading.Lock()
        self.index_array = None
        self.index_generator = self._flow_index()

    def _set_index_array(self):
        self.index_array = np.arange(self.n)
        if self.shuffle:
            self.index_array = np.random.permutation(self.n)

    def __getitem__(self, idx):
        if idx >= len(self):
            raise ValueError(
                "Asked to retrieve element {idx}, "
                "but the Sequence "
                "has length {length}".format(idx=idx, length=len(self))
            )
        if self.seed is not None:
            np.random.seed(self.seed + self.total_batches_seen)
        self.total_batches_seen += 1
        if self.index_array is None:
            self._set_index_array()
        index_array = self.index_array[
            self.batch_size * idx : self.batch_size * (idx + 1)
        ]
        return self._get_batches_of_transformed_samples(index_array)

    def __len__(self):
        return (self.n + self.batch_size - 1) // self.batch_size  # round up

    def on_epoch_end(self):
        self._set_index_array()

    def reset(self):
        self.batch_index = 0

    def _flow_index(self):
        # Ensure self.batch_index is 0.
        self.reset()
        while 1:
            if self.seed is not None:
                np.random.seed(self.seed + self.total_batches_seen)
            if self.batch_index == 0:
                self._set_index_array()

            if self.n == 0:
                # Avoiding modulo by zero error
                current_index = 0
            else:
                current_index = (self.batch_index * self.batch_size) % self.n
            if self.n > current_index + self.batch_size:
                self.batch_index += 1
            else:
                self.batch_index = 0
            self.total_batches_seen += 1
            yield self.index_array[current_index : current_index + self.batch_size]

    def __iter__(self):
        # Needed if we want to do something like:
        # for x, y in data_gen.flow(...):
        return self

    def __next__(self):
        with self.lock:
            index_array = next(self.index_generator)
        # The transformation of images is not under thread lock
        # so it can be done in parallel
        return self._get_batches_of_transformed_samples(index_array)

    def _get_batches_of_transformed_samples(self, index_array):
        """Gets a batch of transformed samples.

        Args:
            index_array: Array of sample indices to include in batch.
        Returns:
            A batch of transformed samples.
        """
        raise NotImplementedError


def _iter_valid_files(directory, white_list_formats, follow_links):
    """Iterates on files with extension.

    Args:
        directory: Absolute path to the directory
            containing files to be counted
        white_list_formats: Set of strings containing allowed extensions for
            the files to be counted.
        follow_links: Boolean, follow symbolic links to subdirectories.
    Yields:
        Tuple of (root, filename) with extension in `white_list_formats`.
    """

    def _recursive_list(subpath):
        return sorted(os.walk(subpath, followlinks=follow_links), key=lambda x: x[0])

    for root, _, files in _recursive_list(directory):
        for fname in sorted(files):
            if fname.lower().endswith(".tiff"):
                warnings.warn(
                    'Using ".tiff" files with multiple bands '
                    "will cause distortion. Please verify your output."
                )
            if fname.lower().endswith(white_list_formats):
                yield root, fname


def _list_valid_filenames_in_directory(
    directory, white_list_formats, split, class_indices, follow_links
):
    """Lists paths of files in `subdir` with extensions in `white_list_formats`.

    Args:
        directory: absolute path to a directory containing the files to list.
            The directory name is used as class label
            and must be a key of `class_indices`.
        white_list_formats: set of strings containing allowed extensions for
            the files to be counted.
        split: tuple of floats (e.g. `(0.2, 0.6)`) to only take into
            account a certain fraction of files in each directory.
            E.g.: `segment=(0.6, 1.0)` would only account for last 40 percent
            of images in each directory.
        class_indices: dictionary mapping a class name to its index.
        follow_links: boolean, follow symbolic links to subdirectories.

    Returns:
         classes: a list of class indices
         filenames: the path of valid files in `directory`, relative from
             `directory`'s parent (e.g., if `directory` is "dataset/class1",
            the filenames will be
            `["class1/file1.jpg", "class1/file2.jpg", ...]`).
    """
    dirname = os.path.basename(directory)
    if split:
        all_files = list(_iter_valid_files(directory, white_list_formats, follow_links))
        num_files = len(all_files)
        start, stop = int(split[0] * num_files), int(split[1] * num_files)
        valid_files = all_files[start:stop]
    else:
        valid_files = _iter_valid_files(directory, white_list_formats, follow_links)
    classes = []
    filenames = []
    for root, fname in valid_files:
        classes.append(class_indices[dirname])
        absolute_path = os.path.join(root, fname)
        relative_path = os.path.join(dirname, os.path.relpath(absolute_path, directory))
        filenames.append(relative_path)

    return classes, filenames


class BatchFromFilesMixin:
    """Adds methods related to getting batches from filenames.

    It includes the logic to transform image files to batches.
    """

    def set_processing_attrs(
        self,
        image_data_generator,
        target_size,
        color_mode,
        data_format,
        save_to_dir,
        save_prefix,
        save_format,
        subset,
        interpolation,
        keep_aspect_ratio,
    ):
        """Sets attributes to use later for processing files into a batch.

        Args:
            image_data_generator: Instance of `ImageDataGenerator`
                to use for random transformations and normalization.
            target_size: tuple of integers, dimensions to resize input images
            to.
            color_mode: One of `"rgb"`, `"rgba"`, `"grayscale"`.
                Color mode to read images.
            data_format: String, one of `channels_first`, `channels_last`.
            save_to_dir: Optional directory where to save the pictures
                being yielded, in a viewable format. This is useful
                for visualizing the random transformations being
                applied, for debugging purposes.
            save_prefix: String prefix to use for saving sample
                images (if `save_to_dir` is set).
            save_format: Format to use for saving sample images
                (if `save_to_dir` is set).
            subset: Subset of data (`"training"` or `"validation"`) if
                validation_split is set in ImageDataGenerator.
            interpolation: Interpolation method used to resample the image if
                the target size is different from that of the loaded image.
                Supported methods are "nearest", "bilinear", and "bicubic". If
                PIL version 1.1.3 or newer is installed, "lanczos" is also
                supported. If PIL version 3.4.0 or newer is installed, "box" and
                "hamming" are also supported. By default, "nearest" is used.
            keep_aspect_ratio: Boolean, whether to resize images to a target
                size without aspect ratio distortion. The image is cropped in
                the center with target aspect ratio before resizing.
        """
        self.image_data_generator = image_data_generator
        self.target_size = tuple(target_size)
        self.keep_aspect_ratio = keep_aspect_ratio
        if color_mode not in {"rgb", "rgba", "grayscale"}:
            raise ValueError(
                f"Invalid color mode: {color_mode}"
                '; expected "rgb", "rgba", or "grayscale".'
            )
        self.color_mode = color_mode
        self.data_format = data_format
        if self.color_mode == "rgba":
            if self.data_format == "channels_last":
                self.image_shape = self.target_size + (4,)
            else:
                self.image_shape = (4,) + self.target_size
        elif self.color_mode == "rgb":
            if self.data_format == "channels_last":
                self.image_shape = self.target_size + (3,)
            else:
                self.image_shape = (3,) + self.target_size
        else:
            if self.data_format == "channels_last":
                self.image_shape = self.target_size + (1,)
            else:
                self.image_shape = (1,) + self.target_size
        self.save_to_dir = save_to_dir
        self.save_prefix = save_prefix
        self.save_format = save_format
        self.interpolation = interpolation
        if subset is not None:
            validation_split = self.image_data_generator._validation_split
            if subset == "validation":
                split = (0, validation_split)
            elif subset == "training":
                split = (validation_split, 1)
            else:
                raise ValueError(
                    f"Invalid subset name: {subset};"
                    'expected "training" or "validation"'
                )
        else:
            split = None
        self.split = split
        self.subset = subset

    def _get_batches_of_transformed_samples(self, index_array):
        """Gets a batch of transformed samples.

        Args:
            index_array: Array of sample indices to include in batch.
        Returns:
            A batch of transformed samples.
        """
        batch_x = np.zeros((len(index_array),) + self.image_shape, dtype=self.dtype)
        # build batch of image data
        # self.filepaths is dynamic, is better to call it once outside the loop
        filepaths = self.filepaths
        for i, j in enumerate(index_array):
            img = image_utils.load_img(
                filepaths[j],
                color_mode=self.color_mode,
                target_size=self.target_size,
                interpolation=self.interpolation,
                keep_aspect_ratio=self.keep_aspect_ratio,
            )
            x = image_utils.img_to_array(img, data_format=self.data_format)
            # Pillow images should be closed after `load_img`,
            # but not PIL images.
            if hasattr(img, "close"):
                img.close()
            if self.image_data_generator:
                params = self.image_data_generator.get_random_transform(x.shape)
                x = self.image_data_generator.apply_transform(x, params)
                x = self.image_data_generator.standardize(x)
            batch_x[i] = x
        # optionally save augmented images to disk for debugging purposes
        if self.save_to_dir:
            for i, j in enumerate(index_array):
                img = image_utils.array_to_img(batch_x[i], self.data_format, scale=True)
                fname = "{prefix}_{index}_{hash}.{format}".format(
                    prefix=self.save_prefix,
                    index=j,
                    hash=np.random.randint(1e7),
                    format=self.save_format,
                )
                img.save(os.path.join(self.save_to_dir, fname))
        # build batch of labels
        if self.class_mode == "input":
            batch_y = batch_x.copy()
        elif self.class_mode in {"binary", "sparse"}:
            batch_y = np.empty(len(batch_x), dtype=self.dtype)
            for i, n_observation in enumerate(index_array):
                batch_y[i] = self.classes[n_observation]
        elif self.class_mode == "categorical":
            batch_y = np.zeros(
                (len(batch_x), len(self.class_indices)), dtype=self.dtype
            )
            for i, n_observation in enumerate(index_array):
                batch_y[i, self.classes[n_observation]] = 1.0
        elif self.class_mode == "multi_output":
            batch_y = [output[index_array] for output in self.labels]
        elif self.class_mode == "raw":
            batch_y = self.labels[index_array]
        else:
            return batch_x
        if self.sample_weight is None:
            return batch_x, batch_y
        else:
            return batch_x, batch_y, self.sample_weight[index_array]

    @property
    def filepaths(self):
        """List of absolute paths to image files."""
        raise NotImplementedError(
            "`filepaths` property method has not "
            "been implemented in {}.".format(type(self).__name__)
        )

    @property
    def labels(self):
        """Class labels of every observation."""
        raise NotImplementedError(
            "`labels` property method has not been implemented in {}.".format(
                type(self).__name__
            )
        )

    @property
    def sample_weight(self):
        raise NotImplementedError(
            "`sample_weight` property method has not "
            "been implemented in {}.".format(type(self).__name__)
        )


@keras_export("keras._legacy.preprocessing.image.DirectoryIterator")
class DirectoryIterator(BatchFromFilesMixin, Iterator):
    """Iterator capable of reading images from a directory on disk.

    DEPRECATED.
    """

    allowed_class_modes = {"categorical", "binary", "sparse", "input", None}

    def __init__(
        self,
        directory,
        image_data_generator,
        target_size=(256, 256),
        color_mode="rgb",
        classes=None,
        class_mode="categorical",
        batch_size=32,
        shuffle=True,
        seed=None,
        data_format=None,
        save_to_dir=None,
        save_prefix="",
        save_format="png",
        follow_links=False,
        subset=None,
        interpolation="nearest",
        keep_aspect_ratio=False,
        dtype=None,
    ):
        if data_format is None:
            data_format = backend.image_data_format()
        if dtype is None:
            dtype = backend.floatx()
        super().set_processing_attrs(
            image_data_generator,
            target_size,
            color_mode,
            data_format,
            save_to_dir,
            save_prefix,
            save_format,
            subset,
            interpolation,
            keep_aspect_ratio,
        )
        self.directory = directory
        self.classes = classes
        if class_mode not in self.allowed_class_modes:
            raise ValueError(
                "Invalid class_mode: {}; expected one of: {}".format(
                    class_mode, self.allowed_class_modes
                )
            )
        self.class_mode = class_mode
        self.dtype = dtype
        # First, count the number of samples and classes.
        self.samples = 0

        if not classes:
            classes = []
            for subdir in sorted(os.listdir(directory)):
                if os.path.isdir(os.path.join(directory, subdir)):
                    classes.append(subdir)
        self.num_classes = len(classes)
        self.class_indices = dict(zip(classes, range(len(classes))))

        pool = multiprocessing.pool.ThreadPool()

        # Second, build an index of the images
        # in the different class subfolders.
        results = []
        self.filenames = []
        i = 0
        for dirpath in (os.path.join(directory, subdir) for subdir in classes):
            results.append(
                pool.apply_async(
                    _list_valid_filenames_in_directory,
                    (
                        dirpath,
                        self.white_list_formats,
                        self.split,
                        self.class_indices,
                        follow_links,
                    ),
                )
            )
        classes_list = []
        for res in results:
            classes, filenames = res.get()
            classes_list.append(classes)
            self.filenames += filenames
        self.samples = len(self.filenames)
        self.classes = np.zeros((self.samples,), dtype="int32")
        for classes in classes_list:
            self.classes[i : i + len(classes)] = classes
            i += len(classes)

        io_utils.print_msg(
            f"Found {self.samples} images belonging to " f"{self.num_classes} classes."
        )
        pool.close()
        pool.join()
        self._filepaths = [
            os.path.join(self.directory, fname) for fname in self.filenames
        ]
        super().__init__(self.samples, batch_size, shuffle, seed)

    @property
    def filepaths(self):
        return self._filepaths

    @property
    def labels(self):
        return self.classes

    @property  # mixin needs this property to work
    def sample_weight(self):
        # no sample weights will be returned
        return None


@keras_export("keras._legacy.preprocessing.image.NumpyArrayIterator")
class NumpyArrayIterator(Iterator):
    """Iterator yielding data from a Numpy array.

    DEPRECATED.
    """

    def __init__(
        self,
        x,
        y,
        image_data_generator,
        batch_size=32,
        shuffle=False,
        sample_weight=None,
        seed=None,
        data_format=None,
        save_to_dir=None,
        save_prefix="",
        save_format="png",
        subset=None,
        ignore_class_split=False,
        dtype=None,
    ):
        if data_format is None:
            data_format = backend.image_data_format()
        if dtype is None:
            dtype = backend.floatx()
        self.dtype = dtype
        if isinstance(x, tuple) or isinstance(x, list):
            if not isinstance(x[1], list):
                x_misc = [np.asarray(x[1])]
            else:
                x_misc = [np.asarray(xx) for xx in x[1]]
            x = x[0]
            for xx in x_misc:
                if len(x) != len(xx):
                    raise ValueError(
                        "All of the arrays in `x` "
                        "should have the same length. "
                        "Found a pair with: "
                        f"len(x[0]) = {len(x)}, len(x[?]) = {len(xx)}"
                    )
        else:
            x_misc = []

        if y is not None and len(x) != len(y):
            raise ValueError(
                "`x` (images tensor) and `y` (labels) "
                "should have the same length. "
                f"Found: x.shape = {np.asarray(x).shape}, "
                f"y.shape = {np.asarray(y).shape}"
            )
        if sample_weight is not None and len(x) != len(sample_weight):
            raise ValueError(
                "`x` (images tensor) and `sample_weight` "
                "should have the same length. "
                f"Found: x.shape = {np.asarray(x).shape}, "
                f"sample_weight.shape = {np.asarray(sample_weight).shape}"
            )
        if subset is not None:
            if subset not in {"training", "validation"}:
                raise ValueError(
                    f"Invalid subset name: {subset}"
                    '; expected "training" or "validation".'
                )
            split_idx = int(len(x) * image_data_generator._validation_split)

            if (
                y is not None
                and not ignore_class_split
                and not np.array_equal(
                    np.unique(y[:split_idx]), np.unique(y[split_idx:])
                )
            ):
                raise ValueError(
                    "Training and validation subsets "
                    "have different number of classes after "
                    "the split. If your numpy arrays are "
                    "sorted by the label, you might want "
                    "to shuffle them."
                )

            if subset == "validation":
                x = x[:split_idx]
                x_misc = [np.asarray(xx[:split_idx]) for xx in x_misc]
                if y is not None:
                    y = y[:split_idx]
            else:
                x = x[split_idx:]
                x_misc = [np.asarray(xx[split_idx:]) for xx in x_misc]
                if y is not None:
                    y = y[split_idx:]

        self.x = np.asarray(x, dtype=self.dtype)
        self.x_misc = x_misc
        if self.x.ndim != 4:
            raise ValueError(
                "Input data in `NumpyArrayIterator` "
                "should have rank 4. You passed an array "
                f"with shape {self.x.shape}"
            )
        channels_axis = 3 if data_format == "channels_last" else 1
        if self.x.shape[channels_axis] not in {1, 3, 4}:
            warnings.warn(
                'NumpyArrayIterator is set to use the data format convention "'
                + data_format
                + '" (channels on axis '
                + str(channels_axis)
                + "), i.e. expected either 1, 3, or 4 channels on axis "
                + str(channels_axis)
                + ". However, it was passed an array with shape "
                + str(self.x.shape)
                + " ("
                + str(self.x.shape[channels_axis])
                + " channels)."
            )
        if y is not None:
            self.y = np.asarray(y)
        else:
            self.y = None
        if sample_weight is not None:
            self.sample_weight = np.asarray(sample_weight)
        else:
            self.sample_weight = None
        self.image_data_generator = image_data_generator
        self.data_format = data_format
        self.save_to_dir = save_to_dir
        self.save_prefix = save_prefix
        self.save_format = save_format
        super().__init__(x.shape[0], batch_size, shuffle, seed)

    def _get_batches_of_transformed_samples(self, index_array):
        batch_x = np.zeros(
            tuple([len(index_array)] + list(self.x.shape)[1:]), dtype=self.dtype
        )
        for i, j in enumerate(index_array):
            x = self.x[j]
            params = self.image_data_generator.get_random_transform(x.shape)
            x = self.image_data_generator.apply_transform(x.astype(self.dtype), params)
            x = self.image_data_generator.standardize(x)
            batch_x[i] = x

        if self.save_to_dir:
            for i, j in enumerate(index_array):
                img = image_utils.array_to_img(batch_x[i], self.data_format, scale=True)
                fname = "{prefix}_{index}_{hash}.{format}".format(
                    prefix=self.save_prefix,
                    index=j,
                    hash=np.random.randint(1e4),
                    format=self.save_format,
                )
                img.save(os.path.join(self.save_to_dir, fname))
        batch_x_miscs = [xx[index_array] for xx in self.x_misc]
        output = (batch_x if not batch_x_miscs else [batch_x] + batch_x_miscs,)
        if self.y is None:
            return output[0]
        output += (self.y[index_array],)
        if self.sample_weight is not None:
            output += (self.sample_weight[index_array],)
        return output


def validate_filename(filename, white_list_formats):
    """Check if a filename refers to a valid file.

    Args:
        filename: String, absolute path to a file
        white_list_formats: Set, allowed file extensions
    Returns:
        A boolean value indicating if the filename is valid or not
    """
    return filename.lower().endswith(white_list_formats) and os.path.isfile(filename)


class DataFrameIterator(BatchFromFilesMixin, Iterator):
    """Iterator capable of reading images from a directory as a dataframe."""

    allowed_class_modes = {
        "binary",
        "categorical",
        "input",
        "multi_output",
        "raw",
        "sparse",
        None,
    }

    def __init__(
        self,
        dataframe,
        directory=None,
        image_data_generator=None,
        x_col="filename",
        y_col="class",
        weight_col=None,
        target_size=(256, 256),
        color_mode="rgb",
        classes=None,
        class_mode="categorical",
        batch_size=32,
        shuffle=True,
        seed=None,
        data_format="channels_last",
        save_to_dir=None,
        save_prefix="",
        save_format="png",
        subset=None,
        interpolation="nearest",
        keep_aspect_ratio=False,
        dtype="float32",
        validate_filenames=True,
    ):
        super().set_processing_attrs(
            image_data_generator,
            target_size,
            color_mode,
            data_format,
            save_to_dir,
            save_prefix,
            save_format,
            subset,
            interpolation,
            keep_aspect_ratio,
        )
        df = dataframe.copy()
        self.directory = directory or ""
        self.class_mode = class_mode
        self.dtype = dtype
        # check that inputs match the required class_mode
        self._check_params(df, x_col, y_col, weight_col, classes)
        if validate_filenames:  # check which image files are valid and keep them
            df = self._filter_valid_filepaths(df, x_col)
        if class_mode not in ["input", "multi_output", "raw", None]:
            df, classes = self._filter_classes(df, y_col, classes)
            num_classes = len(classes)
            # build an index of all the unique classes
            self.class_indices = dict(zip(classes, range(len(classes))))
        # retrieve only training or validation set
        if self.split:
            num_files = len(df)
            start = int(self.split[0] * num_files)
            stop = int(self.split[1] * num_files)
            df = df.iloc[start:stop, :]
        # get labels for each observation
        if class_mode not in ["input", "multi_output", "raw", None]:
            self.classes = self.get_classes(df, y_col)
        self.filenames = df[x_col].tolist()
        self._sample_weight = df[weight_col].values if weight_col else None

        if class_mode == "multi_output":
            self._targets = [np.array(df[col].tolist()) for col in y_col]
        if class_mode == "raw":
            self._targets = df[y_col].values
        self.samples = len(self.filenames)
        validated_string = "validated" if validate_filenames else "non-validated"
        if class_mode in ["input", "multi_output", "raw", None]:
            io_utils.print_msg(
                f"Found {self.samples} {validated_string} image filenames."
            )
        else:
            io_utils.print_msg(
                f"Found {self.samples} {validated_string} image filenames "
                f"belonging to {num_classes} classes."
            )
        self._filepaths = [
            os.path.join(self.directory, fname) for fname in self.filenames
        ]
        super().__init__(self.samples, batch_size, shuffle, seed)

    def _check_params(self, df, x_col, y_col, weight_col, classes):
        # check class mode is one of the currently supported
        if self.class_mode not in self.allowed_class_modes:
            raise ValueError(
                "Invalid class_mode: {}; expected one of: {}".format(
                    self.class_mode, self.allowed_class_modes
                )
            )
        # check that y_col has several column names if class_mode is
        # multi_output
        if (self.class_mode == "multi_output") and not isinstance(y_col, list):
            raise TypeError(
                'If class_mode="{}", y_col must be a list. Received {}.'.format(
                    self.class_mode, type(y_col).__name__
                )
            )
        # check that filenames/filepaths column values are all strings
        if not all(df[x_col].apply(lambda x: isinstance(x, str))):
            raise TypeError(f"All values in column x_col={x_col} must be strings.")
        # check labels are string if class_mode is binary or sparse
        if self.class_mode in {"binary", "sparse"}:
            if not all(df[y_col].apply(lambda x: isinstance(x, str))):
                raise TypeError(
                    'If class_mode="{}", y_col="{}" column '
                    "values must be strings.".format(self.class_mode, y_col)
                )
        # check that if binary there are only 2 different classes
        if self.class_mode == "binary":
            if classes:
                classes = set(classes)
                if len(classes) != 2:
                    raise ValueError(
                        'If class_mode="binary" there must be 2 '
                        "classes. {} class/es were given.".format(len(classes))
                    )
            elif df[y_col].nunique() != 2:
                raise ValueError(
                    'If class_mode="binary" there must be 2 classes. '
                    "Found {} classes.".format(df[y_col].nunique())
                )
        # check values are string, list or tuple if class_mode is categorical
        if self.class_mode == "categorical":
            types = (str, list, tuple)
            if not all(df[y_col].apply(lambda x: isinstance(x, types))):
                raise TypeError(
                    'If class_mode="{}", y_col="{}" column '
                    "values must be type string, list or tuple.".format(
                        self.class_mode, y_col
                    )
                )
        # raise warning if classes are given but will be unused
        if classes and self.class_mode in {
            "input",
            "multi_output",
            "raw",
            None,
        }:
            warnings.warn(
                '`classes` will be ignored given the class_mode="{}"'.format(
                    self.class_mode
                )
            )
        # check that if weight column that the values are numerical
        if weight_col and not issubclass(df[weight_col].dtype.type, np.number):
            raise TypeError(f"Column weight_col={weight_col} must be numeric.")

    def get_classes(self, df, y_col):
        labels = []
        for label in df[y_col]:
            if isinstance(label, (list, tuple)):
                labels.append([self.class_indices[lbl] for lbl in label])
            else:
                labels.append(self.class_indices[label])
        return labels

    @staticmethod
    def _filter_classes(df, y_col, classes):
        df = df.copy()

        def remove_classes(labels, classes):
            if isinstance(labels, (list, tuple)):
                labels = [cls for cls in labels if cls in classes]
                return labels or None
            elif isinstance(labels, str):
                return labels if labels in classes else None
            else:
                raise TypeError(
                    "Expect string, list or tuple "
                    "but found {} in {} column ".format(type(labels), y_col)
                )

        if classes:
            # prepare for membership lookup
            classes = list(collections.OrderedDict.fromkeys(classes).keys())
            df[y_col] = df[y_col].apply(lambda x: remove_classes(x, classes))
        else:
            classes = set()
            for v in df[y_col]:
                if isinstance(v, (list, tuple)):
                    classes.update(v)
                else:
                    classes.add(v)
            classes = sorted(classes)
        return df.dropna(subset=[y_col]), classes

    def _filter_valid_filepaths(self, df, x_col):
        """Keep only dataframe rows with valid filenames.

        Args:
            df: Pandas dataframe containing filenames in a column
            x_col: string, column in `df` that contains the filenames or
                filepaths
        Returns:
            absolute paths to image files
        """
        filepaths = df[x_col].map(lambda fname: os.path.join(self.directory, fname))
        mask = filepaths.apply(validate_filename, args=(self.white_list_formats,))
        n_invalid = (~mask).sum()
        if n_invalid:
            warnings.warn(
                'Found {} invalid image filename(s) in x_col="{}". '
                "These filename(s) will be ignored.".format(n_invalid, x_col)
            )
        return df[mask]

    @property
    def filepaths(self):
        return self._filepaths

    @property
    def labels(self):
        if self.class_mode in {"multi_output", "raw"}:
            return self._targets
        else:
            return self.classes

    @property
    def sample_weight(self):
        return self._sample_weight


def flip_axis(x, axis):
    x = np.asarray(x).swapaxes(axis, 0)
    x = x[::-1, ...]
    x = x.swapaxes(0, axis)
    return x


@keras_export("keras._legacy.preprocessing.image.ImageDataGenerator")
class ImageDataGenerator:
    """DEPRECATED."""

    def __init__(
        self,
        featurewise_center=False,
        samplewise_center=False,
        featurewise_std_normalization=False,
        samplewise_std_normalization=False,
        zca_whitening=False,
        zca_epsilon=1e-6,
        rotation_range=0,
        width_shift_range=0.0,
        height_shift_range=0.0,
        brightness_range=None,
        shear_range=0.0,
        zoom_range=0.0,
        channel_shift_range=0.0,
        fill_mode="nearest",
        cval=0.0,
        horizontal_flip=False,
        vertical_flip=False,
        rescale=None,
        preprocessing_function=None,
        data_format=None,
        validation_split=0.0,
        interpolation_order=1,
        dtype=None,
    ):
        if data_format is None:
            data_format = backend.image_data_format()
        if dtype is None:
            dtype = backend.floatx()

        self.featurewise_center = featurewise_center
        self.samplewise_center = samplewise_center
        self.featurewise_std_normalization = featurewise_std_normalization
        self.samplewise_std_normalization = samplewise_std_normalization
        self.zca_whitening = zca_whitening
        self.zca_epsilon = zca_epsilon
        self.rotation_range = rotation_range
        self.width_shift_range = width_shift_range
        self.height_shift_range = height_shift_range
        self.shear_range = shear_range
        self.zoom_range = zoom_range
        self.channel_shift_range = channel_shift_range
        self.fill_mode = fill_mode
        self.cval = cval
        self.horizontal_flip = horizontal_flip
        self.vertical_flip = vertical_flip
        self.rescale = rescale
        self.preprocessing_function = preprocessing_function
        self.dtype = dtype
        self.interpolation_order = interpolation_order

        if data_format not in {"channels_last", "channels_first"}:
            raise ValueError(
                '`data_format` should be `"channels_last"` '
                "(channel after row and column) or "
                '`"channels_first"` (channel before row and column). '
                f"Received: {data_format}"
            )
        self.data_format = data_format
        if data_format == "channels_first":
            self.channel_axis = 1
            self.row_axis = 2
            self.col_axis = 3
        if data_format == "channels_last":
            self.channel_axis = 3
            self.row_axis = 1
            self.col_axis = 2
        if validation_split and not 0 < validation_split < 1:
            raise ValueError(
                "`validation_split` must be strictly between 0 and 1. "
                f" Received: {validation_split}"
            )
        self._validation_split = validation_split

        self.mean = None
        self.std = None
        self.zca_whitening_matrix = None

        if isinstance(zoom_range, (float, int)):
            self.zoom_range = [1 - zoom_range, 1 + zoom_range]
        elif len(zoom_range) == 2 and all(
            isinstance(val, (float, int)) for val in zoom_range
        ):
            self.zoom_range = [zoom_range[0], zoom_range[1]]
        else:
            raise ValueError(
                "`zoom_range` should be a float or "
                "a tuple or list of two floats. "
                f"Received: {zoom_range}"
            )
        if zca_whitening:
            if not featurewise_center:
                self.featurewise_center = True
                warnings.warn(
                    "This ImageDataGenerator specifies "
                    "`zca_whitening`, which overrides "
                    "setting of `featurewise_center`."
                )
            if featurewise_std_normalization:
                self.featurewise_std_normalization = False
                warnings.warn(
                    "This ImageDataGenerator specifies "
                    "`zca_whitening` "
                    "which overrides setting of"
                    "`featurewise_std_normalization`."
                )
        if featurewise_std_normalization:
            if not featurewise_center:
                self.featurewise_center = True
                warnings.warn(
                    "This ImageDataGenerator specifies "
                    "`featurewise_std_normalization`, "
                    "which overrides setting of "
                    "`featurewise_center`."
                )
        if samplewise_std_normalization:
            if not samplewise_center:
                self.samplewise_center = True
                warnings.warn(
                    "This ImageDataGenerator specifies "
                    "`samplewise_std_normalization`, "
                    "which overrides setting of "
                    "`samplewise_center`."
                )
        if brightness_range is not None:
            if (
                not isinstance(brightness_range, (tuple, list))
                or len(brightness_range) != 2
            ):
                raise ValueError(
                    "`brightness_range should be tuple or list of two floats. "
                    f"Received: {brightness_range}"
                )
        self.brightness_range = brightness_range

    def transform(self, X):
        X_ = np.array([X]) if X.ndim == 3 else X
        batch_x = np.zeros(X_.shape, dtype="float32")
        for i, x in enumerate(X_):
            params = self.get_random_transform(x.shape)
            x = self.apply_transform(x.astype(self.dtype), params)
            x = self.standardize(x)
            batch_x[i] = x
        # batch_x = batch_x[0] if X.ndim==3 else batch_x
        return batch_x

    def flow(
        self,
        x,
        y=None,
        batch_size=32,
        shuffle=True,
        sample_weight=None,
        seed=None,
        save_to_dir=None,
        save_prefix="",
        save_format="png",
        ignore_class_split=False,
        subset=None,
    ):
        return NumpyArrayIterator(
            x,
            y,
            self,
            batch_size=batch_size,
            shuffle=shuffle,
            sample_weight=sample_weight,
            seed=seed,
            data_format=self.data_format,
            save_to_dir=save_to_dir,
            save_prefix=save_prefix,
            save_format=save_format,
            ignore_class_split=ignore_class_split,
            subset=subset,
            dtype=self.dtype,
        )

    def flow_from_directory(
        self,
        directory,
        target_size=(256, 256),
        color_mode="rgb",
        classes=None,
        class_mode="categorical",
        batch_size=32,
        shuffle=True,
        seed=None,
        save_to_dir=None,
        save_prefix="",
        save_format="png",
        follow_links=False,
        subset=None,
        interpolation="nearest",
        keep_aspect_ratio=False,
    ):
        return DirectoryIterator(
            directory,
            self,
            target_size=target_size,
            color_mode=color_mode,
            keep_aspect_ratio=keep_aspect_ratio,
            classes=classes,
            class_mode=class_mode,
            data_format=self.data_format,
            batch_size=batch_size,
            shuffle=shuffle,
            seed=seed,
            save_to_dir=save_to_dir,
            save_prefix=save_prefix,
            save_format=save_format,
            follow_links=follow_links,
            subset=subset,
            interpolation=interpolation,
            dtype=self.dtype,
        )

    def flow_from_dataframe(
        self,
        dataframe,
        directory=None,
        x_col="filename",
        y_col="class",
        weight_col=None,
        target_size=(256, 256),
        color_mode="rgb",
        classes=None,
        class_mode="categorical",
        batch_size=32,
        shuffle=True,
        seed=None,
        save_to_dir=None,
        save_prefix="",
        save_format="png",
        subset=None,
        interpolation="nearest",
        validate_filenames=True,
        **kwargs,
    ):
        if "has_ext" in kwargs:
            warnings.warn(
                "has_ext is deprecated, filenames in the dataframe have "
                "to match the exact filenames in disk.",
                DeprecationWarning,
            )
        if "sort" in kwargs:
            warnings.warn(
                "sort is deprecated, batches will be created in the"
                "same order than the filenames provided if `shuffle`"
                "is set to `False`.",
                DeprecationWarning,
            )
        if class_mode == "other":
            warnings.warn(
                '`class_mode="other"` is deprecated, please use ' '`class_mode="raw"`.',
                DeprecationWarning,
            )
            class_mode = "raw"
        if "drop_duplicates" in kwargs:
            warnings.warn(
                "drop_duplicates is deprecated, you can drop duplicates "
                "by using the pandas.DataFrame.drop_duplicates method.",
                DeprecationWarning,
            )

        return DataFrameIterator(
            dataframe,
            directory,
            self,
            x_col=x_col,
            y_col=y_col,
            weight_col=weight_col,
            target_size=target_size,
            color_mode=color_mode,
            classes=classes,
            class_mode=class_mode,
            data_format=self.data_format,
            batch_size=batch_size,
            shuffle=shuffle,
            seed=seed,
            save_to_dir=save_to_dir,
            save_prefix=save_prefix,
            save_format=save_format,
            subset=subset,
            interpolation=interpolation,
            validate_filenames=validate_filenames,
            dtype=self.dtype,
        )

    def standardize(self, x):
        """Applies the normalization configuration in-place to a batch of
        inputs.

        `x` is changed in-place since the function is mainly used internally
        to standardize images and feed them to your network. If a copy of `x`
        would be created instead it would have a significant performance cost.
        If you want to apply this method without changing the input in-place
        you can call the method creating a copy before:

        standardize(np.copy(x))

        Args:
            x: Batch of inputs to be normalized.

        Returns:
            The inputs, normalized.
        """
        if self.preprocessing_function:
            x = self.preprocessing_function(x)
        if self.rescale:
            x *= self.rescale
        if self.samplewise_center:
            x -= np.mean(x, keepdims=True)
        if self.samplewise_std_normalization:
            x /= np.std(x, keepdims=True) + 1e-6

        if self.featurewise_center:
            if self.mean is not None:
                x -= self.mean
            else:
                warnings.warn(
                    "This ImageDataGenerator specifies "
                    "`featurewise_center`, but it hasn't "
                    "been fit on any training data. Fit it "
                    "first by calling `.fit(numpy_data)`."
                )
        if self.featurewise_std_normalization:
            if self.std is not None:
                x /= self.std + 1e-6
            else:
                warnings.warn(
                    "This ImageDataGenerator specifies "
                    "`featurewise_std_normalization`, "
                    "but it hasn't "
                    "been fit on any training data. Fit it "
                    "first by calling `.fit(numpy_data)`."
                )
        if self.zca_whitening:
            if self.zca_whitening_matrix is not None:
                flat_x = x.reshape(-1, np.prod(x.shape[-3:]))
                white_x = flat_x @ self.zca_whitening_matrix
                x = np.reshape(white_x, x.shape)
            else:
                warnings.warn(
                    "This ImageDataGenerator specifies "
                    "`zca_whitening`, but it hasn't "
                    "been fit on any training data. Fit it "
                    "first by calling `.fit(numpy_data)`."
                )
        return x

    def get_random_transform(self, img_shape, seed=None):
        """Generates random parameters for a transformation.

        Args:
            img_shape: Tuple of integers.
                Shape of the image that is transformed.
            seed: Random seed.

        Returns:
            A dictionary containing randomly chosen parameters describing the
            transformation.
        """
        img_row_axis = self.row_axis - 1
        img_col_axis = self.col_axis - 1

        if seed is not None:
            np.random.seed(seed)

        if self.rotation_range:
            theta = np.random.uniform(-self.rotation_range, self.rotation_range)
        else:
            theta = 0

        if self.height_shift_range:
            try:  # 1-D array-like or int
                tx = np.random.choice(self.height_shift_range)
                tx *= np.random.choice([-1, 1])
            except ValueError:  # floating point
                tx = np.random.uniform(
                    -self.height_shift_range, self.height_shift_range
                )
            if np.max(self.height_shift_range) < 1:
                tx *= img_shape[img_row_axis]
        else:
            tx = 0

        if self.width_shift_range:
            try:  # 1-D array-like or int
                ty = np.random.choice(self.width_shift_range)
                ty *= np.random.choice([-1, 1])
            except ValueError:  # floating point
                ty = np.random.uniform(-self.width_shift_range, self.width_shift_range)
            if np.max(self.width_shift_range) < 1:
                ty *= img_shape[img_col_axis]
        else:
            ty = 0

        if self.shear_range:
            shear = np.random.uniform(-self.shear_range, self.shear_range)
        else:
            shear = 0

        if self.zoom_range[0] == 1 and self.zoom_range[1] == 1:
            zx, zy = 1, 1
        else:
            zx, zy = np.random.uniform(self.zoom_range[0], self.zoom_range[1], 2)

        flip_horizontal = (np.random.random() < 0.5) * self.horizontal_flip
        flip_vertical = (np.random.random() < 0.5) * self.vertical_flip

        channel_shift_intensity = None
        if self.channel_shift_range != 0:
            channel_shift_intensity = np.random.uniform(
                -self.channel_shift_range, self.channel_shift_range
            )

        brightness = None
        if self.brightness_range is not None:
            brightness = np.random.uniform(
                self.brightness_range[0], self.brightness_range[1]
            )

        transform_parameters = {
            "theta": theta,
            "tx": tx,
            "ty": ty,
            "shear": shear,
            "zx": zx,
            "zy": zy,
            "flip_horizontal": flip_horizontal,
            "flip_vertical": flip_vertical,
            "channel_shift_intensity": channel_shift_intensity,
            "brightness": brightness,
        }

        return transform_parameters

    def apply_transform(self, x, transform_parameters):
        """Applies a transformation to an image according to given parameters.

        Args:
            x: 3D tensor, single image.
            transform_parameters: Dictionary with string - parameter pairs
                describing the transformation.
                Currently, the following parameters
                from the dictionary are used:
                - `'theta'`: Float. Rotation angle in degrees.
                - `'tx'`: Float. Shift in the x direction.
                - `'ty'`: Float. Shift in the y direction.
                - `'shear'`: Float. Shear angle in degrees.
                - `'zx'`: Float. Zoom in the x direction.
                - `'zy'`: Float. Zoom in the y direction.
                - `'flip_horizontal'`: Boolean. Horizontal flip.
                - `'flip_vertical'`: Boolean. Vertical flip.
                - `'channel_shift_intensity'`: Float. Channel shift intensity.
                - `'brightness'`: Float. Brightness shift intensity.

        Returns:
            A transformed version of the input (same shape).
        """
        # x is a single image, so it doesn't have image number at index 0
        img_row_axis = self.row_axis - 1
        img_col_axis = self.col_axis - 1
        img_channel_axis = self.channel_axis - 1

        x = apply_affine_transform(
            x,
            transform_parameters.get("theta", 0),
            transform_parameters.get("tx", 0),
            transform_parameters.get("ty", 0),
            transform_parameters.get("shear", 0),
            transform_parameters.get("zx", 1),
            transform_parameters.get("zy", 1),
            row_axis=img_row_axis,
            col_axis=img_col_axis,
            channel_axis=img_channel_axis,
            fill_mode=self.fill_mode,
            cval=self.cval,
            order=self.interpolation_order,
        )

        if transform_parameters.get("channel_shift_intensity") is not None:
            x = apply_channel_shift(
                x,
                transform_parameters["channel_shift_intensity"],
                img_channel_axis,
            )

        if transform_parameters.get("flip_horizontal", False):
            x = flip_axis(x, img_col_axis)

        if transform_parameters.get("flip_vertical", False):
            x = flip_axis(x, img_row_axis)

        if transform_parameters.get("brightness") is not None:
            x = apply_brightness_shift(x, transform_parameters["brightness"], False)

        return x

    def random_transform(self, x, seed=None):
        """Applies a random transformation to an image.

        Args:
            x: 3D tensor, single image.
            seed: Random seed.

        Returns:
            A randomly transformed version of the input (same shape).
        """
        params = self.get_random_transform(x.shape, seed)
        return self.apply_transform(x, params)

    def fit(self, x, augment=False, rounds=1, seed=None):
        """Fits the data generator to some sample data.

        This computes the internal data stats related to the
        data-dependent transformations, based on an array of sample data.

        Only required if `featurewise_center` or
        `featurewise_std_normalization` or `zca_whitening`
        are set to `True`.

        When `rescale` is set to a value, rescaling is applied to
        sample data before computing the internal data stats.

        Args:
            x: Sample data. Should have rank 4.
             In case of grayscale data,
             the channels axis should have value 1, in case
             of RGB data, it should have value 3, and in case
             of RGBA data, it should have value 4.
            augment: Boolean (default: False).
                Whether to fit on randomly augmented samples.
            rounds: Int (default: 1).
                If using data augmentation (`augment=True`),
                this is how many augmentation passes over the data to use.
            seed: Int (default: None). Random seed.
        """
        x = np.asarray(x, dtype=self.dtype)
        if x.ndim != 4:
            raise ValueError(
                "Input to `.fit()` should have rank 4. Got array with shape: "
                + str(x.shape)
            )
        if x.shape[self.channel_axis] not in {1, 3, 4}:
            warnings.warn(
                "Expected input to be images (as Numpy array) "
                'following the data format convention "'
                + self.data_format
                + '" (channels on axis '
                + str(self.channel_axis)
                + "), i.e. expected either 1, 3 or 4 channels on axis "
                + str(self.channel_axis)
                + ". However, it was passed an array with shape "
                + str(x.shape)
                + " ("
                + str(x.shape[self.channel_axis])
                + " channels)."
            )

        if seed is not None:
            np.random.seed(seed)

        x = np.copy(x)
        if self.rescale:
            x *= self.rescale

        if augment:
            ax = np.zeros(
                tuple([rounds * x.shape[0]] + list(x.shape)[1:]),
                dtype=self.dtype,
            )
            for r in range(rounds):
                for i in range(x.shape[0]):
                    ax[i + r * x.shape[0]] = self.random_transform(x[i])
            x = ax

        if self.featurewise_center:
            self.mean = np.mean(x, axis=(0, self.row_axis, self.col_axis))
            broadcast_shape = [1, 1, 1]
            broadcast_shape[self.channel_axis - 1] = x.shape[self.channel_axis]
            self.mean = np.reshape(self.mean, broadcast_shape)
            x -= self.mean

        if self.featurewise_std_normalization:
            self.std = np.std(x, axis=(0, self.row_axis, self.col_axis))
            broadcast_shape = [1, 1, 1]
            broadcast_shape[self.channel_axis - 1] = x.shape[self.channel_axis]
            self.std = np.reshape(self.std, broadcast_shape)
            x /= self.std + 1e-6

        if self.zca_whitening:
            n = len(x)
            flat_x = np.reshape(x, (n, -1))

            u, s, _ = np.linalg.svd(flat_x.T, full_matrices=False)
            s_inv = np.sqrt(n) / (s + self.zca_epsilon)
            self.zca_whitening_matrix = (u * s_inv).dot(u.T)


@keras_export("keras._legacy.preprocessing.image.random_rotation")
def random_rotation(
    x,
    rg,
    row_axis=1,
    col_axis=2,
    channel_axis=0,
    fill_mode="nearest",
    cval=0.0,
    interpolation_order=1,
):
    """DEPRECATED."""
    theta = np.random.uniform(-rg, rg)
    x = apply_affine_transform(
        x,
        theta=theta,
        row_axis=row_axis,
        col_axis=col_axis,
        channel_axis=channel_axis,
        fill_mode=fill_mode,
        cval=cval,
        order=interpolation_order,
    )
    return x


@keras_export("keras._legacy.preprocessing.image.random_shift")
def random_shift(
    x,
    wrg,
    hrg,
    row_axis=1,
    col_axis=2,
    channel_axis=0,
    fill_mode="nearest",
    cval=0.0,
    interpolation_order=1,
):
    """DEPRECATED."""
    h, w = x.shape[row_axis], x.shape[col_axis]
    tx = np.random.uniform(-hrg, hrg) * h
    ty = np.random.uniform(-wrg, wrg) * w
    x = apply_affine_transform(
        x,
        tx=tx,
        ty=ty,
        row_axis=row_axis,
        col_axis=col_axis,
        channel_axis=channel_axis,
        fill_mode=fill_mode,
        cval=cval,
        order=interpolation_order,
    )
    return x


@keras_export("keras._legacy.preprocessing.image.random_shear")
def random_shear(
    x,
    intensity,
    row_axis=1,
    col_axis=2,
    channel_axis=0,
    fill_mode="nearest",
    cval=0.0,
    interpolation_order=1,
):
    """DEPRECATED."""
    shear = np.random.uniform(-intensity, intensity)
    x = apply_affine_transform(
        x,
        shear=shear,
        row_axis=row_axis,
        col_axis=col_axis,
        channel_axis=channel_axis,
        fill_mode=fill_mode,
        cval=cval,
        order=interpolation_order,
    )
    return x


@keras_export("keras._legacy.preprocessing.image.random_zoom")
def random_zoom(
    x,
    zoom_range,
    row_axis=1,
    col_axis=2,
    channel_axis=0,
    fill_mode="nearest",
    cval=0.0,
    interpolation_order=1,
):
    """DEPRECATED."""
    if len(zoom_range) != 2:
        raise ValueError(
            "`zoom_range` should be a tuple or list of two floats. "
            f"Received: {zoom_range}"
        )

    if zoom_range[0] == 1 and zoom_range[1] == 1:
        zx, zy = 1, 1
    else:
        zx, zy = np.random.uniform(zoom_range[0], zoom_range[1], 2)
    x = apply_affine_transform(
        x,
        zx=zx,
        zy=zy,
        row_axis=row_axis,
        col_axis=col_axis,
        channel_axis=channel_axis,
        fill_mode=fill_mode,
        cval=cval,
        order=interpolation_order,
    )
    return x


@keras_export("keras._legacy.preprocessing.image.apply_channel_shift")
def apply_channel_shift(x, intensity, channel_axis=0):
    """Performs a channel shift.

    DEPRECATED.

    Args:
        x: Input tensor. Must be 3D.
        intensity: Transformation intensity.
        channel_axis: Index of axis for channels in the input tensor.

    Returns:
        Numpy image tensor.
    """
    x = np.rollaxis(x, channel_axis, 0)
    min_x, max_x = np.min(x), np.max(x)
    channel_images = [np.clip(x_channel + intensity, min_x, max_x) for x_channel in x]
    x = np.stack(channel_images, axis=0)
    x = np.rollaxis(x, 0, channel_axis + 1)
    return x


@keras_export("keras._legacy.preprocessing.image.random_channel_shift")
def random_channel_shift(x, intensity_range, channel_axis=0):
    """Performs a random channel shift.

    DEPRECATED.

    Args:
        x: Input tensor. Must be 3D.
        intensity_range: Transformation intensity.
        channel_axis: Index of axis for channels in the input tensor.

    Returns:
        Numpy image tensor.
    """
    intensity = np.random.uniform(-intensity_range, intensity_range)
    return apply_channel_shift(x, intensity, channel_axis=channel_axis)


@keras_export("keras._legacy.preprocessing.image.apply_brightness_shift")
def apply_brightness_shift(x, brightness, scale=True):
    """Performs a brightness shift.

    DEPRECATED.

    Args:
        x: Input tensor. Must be 3D.
        brightness: Float. The new brightness value.
        scale: Whether to rescale the image such that minimum and maximum values
            are 0 and 255 respectively. Default: True.

    Returns:
        Numpy image tensor.

    Raises:
        ImportError: if PIL is not available.
    """
    from PIL import ImageEnhance

    x_min, x_max = np.min(x), np.max(x)
    local_scale = (x_min < 0) or (x_max > 255)
    x = image_utils.array_to_img(x, scale=local_scale or scale)
    x = imgenhancer_Brightness = ImageEnhance.Brightness(x)
    x = imgenhancer_Brightness.enhance(brightness)
    x = image_utils.img_to_array(x)
    if not scale and local_scale:
        x = x / 255 * (x_max - x_min) + x_min
    return x


@keras_export("keras._legacy.preprocessing.image.random_brightness")
def random_brightness(x, brightness_range, scale=True):
    """Performs a random brightness shift.

    DEPRECATED.

    Args:
        x: Input tensor. Must be 3D.
        brightness_range: Tuple of floats; brightness range.
        scale: Whether to rescale the image such that minimum and maximum values
            are 0 and 255 respectively. Default: True.

    Returns:
        Numpy image tensor.

    Raises:
        ValueError if `brightness_range` isn't a tuple.
    """
    if len(brightness_range) != 2:
        raise ValueError(
            "`brightness_range should be tuple or list of two floats. "
            f"Received: {brightness_range}"
        )

    u = np.random.uniform(brightness_range[0], brightness_range[1])
    return apply_brightness_shift(x, u, scale)


def transform_matrix_offset_center(matrix, x, y):
    o_x = float(x) / 2 - 0.5
    o_y = float(y) / 2 - 0.5
    offset_matrix = np.array([[1, 0, o_x], [0, 1, o_y], [0, 0, 1]])
    reset_matrix = np.array([[1, 0, -o_x], [0, 1, -o_y], [0, 0, 1]])
    transform_matrix = np.dot(np.dot(offset_matrix, matrix), reset_matrix)
    return transform_matrix


@keras_export("keras._legacy.preprocessing.image.apply_affine_transform")
def apply_affine_transform(
    x,
    theta=0,
    tx=0,
    ty=0,
    shear=0,
    zx=1,
    zy=1,
    row_axis=1,
    col_axis=2,
    channel_axis=0,
    fill_mode="nearest",
    cval=0.0,
    order=1,
):
    """Applies an affine transformation specified by the parameters given.

    DEPRECATED.
    """
    # Input sanity checks:
    # 1. x must 2D image with one or more channels (i.e., a 3D tensor)
    # 2. channels must be either first or last dimension
    if np.unique([row_axis, col_axis, channel_axis]).size != 3:
        raise ValueError("'row_axis', 'col_axis', and 'channel_axis' must be distinct")

    # shall we support negative indices?
    valid_indices = set([0, 1, 2])
    actual_indices = set([row_axis, col_axis, channel_axis])
    if actual_indices != valid_indices:
        raise ValueError(f"Invalid axis' indices: {actual_indices - valid_indices}")

    if x.ndim != 3:
        raise ValueError("Input arrays must be multi-channel 2D images.")
    if channel_axis not in [0, 2]:
        raise ValueError("Channels are allowed and the first and last dimensions.")

    transform_matrix = None
    if theta != 0:
        theta = np.deg2rad(theta)
        rotation_matrix = np.array(
            [
                [np.cos(theta), -np.sin(theta), 0],
                [np.sin(theta), np.cos(theta), 0],
                [0, 0, 1],
            ]
        )
        transform_matrix = rotation_matrix

    if tx != 0 or ty != 0:
        shift_matrix = np.array([[1, 0, tx], [0, 1, ty], [0, 0, 1]])
        if transform_matrix is None:
            transform_matrix = shift_matrix
        else:
            transform_matrix = np.dot(transform_matrix, shift_matrix)

    if shear != 0:
        shear = np.deg2rad(shear)
        shear_matrix = np.array(
            [[1, -np.sin(shear), 0], [0, np.cos(shear), 0], [0, 0, 1]]
        )
        if transform_matrix is None:
            transform_matrix = shear_matrix
        else:
            transform_matrix = np.dot(transform_matrix, shear_matrix)

    if zx != 1 or zy != 1:
        zoom_matrix = np.array([[zx, 0, 0], [0, zy, 0], [0, 0, 1]])
        if transform_matrix is None:
            transform_matrix = zoom_matrix
        else:
            transform_matrix = np.dot(transform_matrix, zoom_matrix)

    if transform_matrix is not None:
        h, w = x.shape[row_axis], x.shape[col_axis]
        transform_matrix = transform_matrix_offset_center(transform_matrix, h, w)
        x = np.rollaxis(x, channel_axis, 0)

        # Matrix construction assumes that coordinates are x, y (in that order).
        # However, regular numpy arrays use y,x (aka i,j) indexing.
        # Possible solution is:
        #   1. Swap the x and y axes.
        #   2. Apply transform.
        #   3. Swap the x and y axes again to restore image-like data ordering.
        # Mathematically, it is equivalent to the following transformation:
        # M' = PMP, where P is the permutation matrix, M is the original
        # transformation matrix.
        if col_axis > row_axis:
            transform_matrix[:, [0, 1]] = transform_matrix[:, [1, 0]]
            transform_matrix[[0, 1]] = transform_matrix[[1, 0]]
        final_affine_matrix = transform_matrix[:2, :2]
        final_offset = transform_matrix[:2, 2]

        channel_images = [
            scipy.ndimage.interpolation.affine_transform(
                x_channel,
                final_affine_matrix,
                final_offset,
                order=order,
                mode=fill_mode,
                cval=cval,
            )
            for x_channel in x
        ]
        x = np.stack(channel_images, axis=0)
        x = np.rollaxis(x, 0, channel_axis + 1)
    return x



import matplotlib.pyplot as plt


def plot_metric(model_training_history, metric_name_1, metric_name_2, plot_name):
    metric_value_1 = model_training_history.history[metric_name_1]
    metric_value_2 = model_training_history.history[metric_name_2]

    epochs = range(len(metric_value_1))

    plt.plot(epochs, metric_value_1, "blue", label=metric_name_1)
    plt.plot(epochs, metric_value_2, "red", label=metric_name_2)

    plt.title(str(plot_name))

    plt.legend()



import os


class CustomOS:
    def __getattr__(self, name):
        return getattr(os, name)

    def listdir(self, dir_pth, sort=False):
        dir_items = os.listdir(dir_pth)
        return sorted(dir_items, key=lambda x: int(x)) if sort else dir_items


custom_os = CustomOS()



import os
import cv2
import numpy as np


def classify_video(input_video_path,
                   preprocess_frame, 
                   model, 
                   labels_dict,
                   default_output_dir="outputs",
                   output_filename=None, 
                   save=False, 
                   show=True
                   ):
    """
    Process a video by writing text on each frame and saving the result as a new video.

    Args:
        input_video_path (str): Path to the input video file.
        preprocess_frame (func): Function to preprocess each frame
        model (keras model): A classification model
        labels_dict (list, dict): A list or dictionary having class names
        default_output_dir (str): An output directory where video is to be saved
        output_filename (str): Output video file's name
        save (bool): True for saving the video otherwise False
        show (bool): True for showing each processed frame otherwise False

    Returns:
        list(tuple): A list of tuples containing prediction info for each frame.
                     (probas, pred_label, pred_class, pred_conf)
    """

    os.makedirs(default_output_dir, exist_ok=True)

    # Open the input video
    cap = cv2.VideoCapture(input_video_path)
    
    width = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
    height = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))

    # Check if video opened successfully
    if not cap.isOpened():
        print("Error: Cannot open video file")
        return

    if save:
        if not output_filename:
            basename = os.path.basename(input_video_path)
            basename_wo_ext, ext = os.path.splitext(basename)
            vid_pth_wo_ext = f"{default_output_dir}/{basename_wo_ext}"
            save_as = f"{vid_pth_wo_ext}_classified{ext}"
        else:
            save_as = f"{default_output_dir}/{output_filename}"

        # Get video properties
        fps = int(cap.get(cv2.CAP_PROP_FPS))

        frame_size = (width, height)
        fourcc = cv2.VideoWriter_fourcc(*'mp4v')  # Codec for .mp4

        # Create VideoWriter object
        out = cv2.VideoWriter(save_as, fourcc, fps, frame_size)

    window_w = 720
    dynamic_scale = int(width/window_w)

    if show:
        cv2.namedWindow("Frame", cv2.WINDOW_NORMAL)  # Allow resizing
        window_h = int(window_w*height/width)
        cv2.resizeWindow("Frame", window_w, window_h)  # Set the window size


    prediction_info = []
    # Process each frame
    while True:
        ret, frame = cap.read()
        if not ret:  # End of video
            break

        # make prediction
        frame_mod = preprocess_frame(frame)
        probas = model.predict(np.expand_dims(frame_mod, 0), verbose=0)[0]
        pred_label = probas.argmax()
        pred_class = labels_dict[pred_label]
        pred_conf = probas[pred_label]

        prediction_info.append((probas, pred_label, pred_class, pred_conf))

        # Write text on the frame
        position = (50, 50*dynamic_scale)  # Text position (x, y)
        font = cv2.FONT_HERSHEY_SIMPLEX
        font_scale = 1*dynamic_scale
        color = (0, 255, 0)  # Green text
        thickness = 2*dynamic_scale

        cv2.putText(frame, pred_class, position, font, font_scale, color, thickness)
        cv2.putText(frame, f"{pred_conf:.2f}", (50, 100*dynamic_scale), font, font_scale, color, thickness)

        # Display the frame
        if show:
            cv2.imshow("Frame", frame)

        key = cv2.waitKey(1)
        if key == ord("q"):  # Press 'q' to exit
            break

        # Write the frame to the output video
        if save: out.write(frame)

    # Release resources
    cap.release()
    if save:
        print(f"Processed video saved at {save_as}")
        out.release()
    if show: cv2.destroyAllWindows()

    return prediction_info


import cv2
import mediapipe as mp
from threading import Thread
from queue import Queue, Empty
import numpy as np
import tempfile
from mlu_tools.validation import validate_file


mp_hands = mp.solutions.hands
mp_drawing = mp.solutions.drawing_utils


def process_frame(input_queue, output_queue):
    hands = mp_hands.Hands()
    while True:
        frame = input_queue.get()
        if frame is None:  # Stop signal
            break

        # Convert frame to RGB and process with MediaPipe
        frame_rgb = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
        results = hands.process(frame_rgb)
        output_queue.put((frame, results))  # Send back the processed frame and results

    hands.close()


def detect_hand_landmarks(source):
    delay = 1
    if type(source) == np.ndarray:
        delay = 0
        with tempfile.NamedTemporaryFile(suffix=".jpg", delete=True) as temp_file:
            cv2.imwrite(temp_file.name, source[..., ::-1])  # Save NumPy array as an image
            cap = cv2.VideoCapture(temp_file.name)
    else:
        if validate_file(source, valid_types=["img"], raise_exception=False):
            delay = 0
        cap = cv2.VideoCapture(source)
    input_queue = Queue(maxsize=1)  # Limit queue size to 1 for the latest frame
    output_queue = Queue(maxsize=1)  # Limit queue size to 1 for the latest results

    # Start the worker thread
    thread = Thread(target=process_frame, args=(input_queue, output_queue), daemon=True)
    thread.start()

    while cap.isOpened():
        ret, frame = cap.read()
        if not ret:
            break

        # Try to enqueue the latest frame (discard older frames if queue is full)
        if not input_queue.full():
            input_queue.put(frame)

        # Render the latest processed frame
        try:
            processed_frame, results = output_queue.get(
                timeout=0.1
            )  # Use timeout to avoid blocking indefinitely
            if results.multi_hand_landmarks:
                for hand_landmarks in results.multi_hand_landmarks:
                    mp_drawing.draw_landmarks(
                        processed_frame, hand_landmarks, mp_hands.HAND_CONNECTIONS
                    )
            # Show the processed frame with hand landmarks
            cv2.imshow("Hand Tracking", processed_frame)
        except Empty:
            pass  # If no result is available, just continue

        # Check if 'q' is pressed for exiting
        if cv2.waitKey(delay) & 0xFF == ord("q"):
            break

        if delay == 0:
            # Render the last processed frame
            try:
                processed_frame, results = output_queue.get(
                    timeout=0.1
                )  # Use timeout to avoid blocking indefinitely
                if results.multi_hand_landmarks:
                    for hand_landmarks in results.multi_hand_landmarks:
                        mp_drawing.draw_landmarks(
                            processed_frame, hand_landmarks, mp_hands.HAND_CONNECTIONS
                        )
                # Show the processed frame with hand landmarks
                cv2.imshow("Hand Tracking", processed_frame)
            except Empty:
                pass  # If no result is available, just continue

            cv2.waitKey(delay)


    # Clean up
    input_queue.put(None)  # Send stop signal to the worker thread
    thread.join()  # Wait for the worker thread to finish
    cap.release()  # Release the webcam
    cv2.destroyAllWindows()  # Ensure the window is properly closed



import random
import tensorflow as tf
import numpy as np
import requests
from tqdm import tqdm
import os
from .data_structure import TreeNode
import gdown
from yt_dlp import YoutubeDL
import zipfile
import tarfile
import cv2
from IPython.display import FileLink
from datetime import datetime
from mega import Mega


def set_global_seed(seed_value):
    # Set random seed for Python's random module
    random.seed(seed_value)

    # Set random seed for NumPy
    np.random.seed(seed_value)

    # Set random seed for TensorFlow
    tf.random.set_seed(seed_value)


def download(file_url, file_save_path, download_from="drive", force=False):
    dest_path = os.path.dirname(file_save_path)
    dest_filename = os.path.basename(file_save_path)
    print(f"Downloading {dest_filename} to {dest_path}/...")
    if os.path.exists(file_save_path) and not force:
        # print("File already exists!\n")
        print(f"{file_save_path} already exists. Skipping download.\n")
        return
    
    if dest_path:
        os.makedirs(dest_path, exist_ok=True)

    if download_from == "web":
        response = requests.get(file_url, stream=True)
        total_size = int(response.headers.get("content-length", 0))

        with open(file_save_path, "wb") as file:
            with tqdm(total=total_size, unit="B", unit_scale=True) as bar:
                for chunk in response.iter_content(chunk_size=1024):
                    file.write(chunk)
                    bar.update(len(chunk))
    elif download_from == "mega":
        # Initialize Mega object
        mega = Mega()
        # Download the file
        file = mega.download_url(file_url, dest_path=(dest_path or "."), 
                        dest_filename=dest_filename)
        print(f"File downloaded at {file}\n")
    else:
        FILE_ID = file_url.split("/")[-2]
        download_url = f"https://drive.google.com/uc?id={FILE_ID}&export=download"
        gdown.download(download_url, file_save_path)


def tree(dir_pth):
    nodes = {}

    for dirpath, dirnames, filenames in sorted(os.walk(dir_pth)):
        if os.path.dirname(dirpath) not in nodes:
            nodes[dirpath] = TreeNode(os.path.basename(dirpath))
        else:
            if len(filenames):
                child_node = TreeNode(f"{os.path.basename(dirpath)} - {len(filenames)}")
            else:
                child_node = TreeNode(f"{os.path.basename(dirpath)}")
            nodes[os.path.dirname(dirpath)].add_child(child_node)
            nodes[dirpath] = child_node

    nodes[dir_pth].display()


def get_dynamic_path(path):
    path_wo_ext, ext = os.path.splitext(path)
    i = 2
    while os.path.exists(path):
        path = f"{path_wo_ext}_{i}{ext}"
        i += 1

    return path


def download_yt_playlist(playlist_url, quality="best"):
    # Configure download options
    format_mapping = {
        "low": "bestvideo[height<=480][ext=mp4]+bestaudio[ext=m4a]/best[height<=480][ext=mp4]",
        "medium": "bestvideo[height<=720][ext=mp4]+bestaudio[ext=m4a]/best[height<=720][ext=mp4]",
        "high": "bestvideo[height<=1080][ext=mp4]+bestaudio[ext=m4a]/best[height<=1080][ext=mp4]",
        "best": "bestvideo[ext=mp4]+bestaudio[ext=m4a]/best[ext=mp4]",
    }

    options = {
        "outtmpl": os.path.join("downloads", "%(title)s.%(ext)s"),
        "format": format_mapping.get(quality, "best"),
        "ignoreerrors": True,  # Ignore errors like private videos
        "no_post_overwrites": True,  # Avoid overwriting the existing files
    }

    # Download the playlist
    with YoutubeDL(options) as ydl:
        ydl.download([playlist_url])

    print(f"Playlist downloaded successfully as {quality} quality MP4!")


def unpack_archive(file_path, target_dir=None, force=False):
    target_dir = target_dir or (os.path.dirname(file_path) if os.path.dirname(file_path) else ".")
    
    # Determine archive type and get root dir
    if file_path.endswith(".zip"):
        with zipfile.ZipFile(file_path, "r") as zip_ref:
            members = zip_ref.namelist()

    elif file_path.endswith((".tar", ".tar.gz", ".tgz")):
        with tarfile.open(file_path, "r") as tar_ref:
            members = tar_ref.getnames()

    else:
        print("Unsupported file type")
        return

    # CHECKING IF ARCHIVE CONTAINS A ROOT FOLDER OR NOT
    archive_contains_a_root_folder = all([os.path.dirname(member) for member in members])
    if archive_contains_a_root_folder:
        root_dir = os.path.dirname(members[0])
    else:
        root_dir = os.path.splitext(os.path.basename(file_path))[0]

    unpacked_dir = os.path.join(target_dir, root_dir)
    print(f"Unpacking {file_path} to {unpacked_dir}...")
    if os.path.exists(unpacked_dir) and not force:
        print(f"{unpacked_dir} already exists. Skipping unpacking.\n")
        return

    if not archive_contains_a_root_folder:
        os.makedirs(unpacked_dir)

    # Unpack with progress bar
    if file_path.endswith(".zip"):
        with zipfile.ZipFile(file_path, "r") as zip_ref:
            for member in tqdm(members, desc="Extracting", unit="file"):
                if archive_contains_a_root_folder:
                    zip_ref.extract(member, target_dir)
                else:
                    zip_ref.extract(member, unpacked_dir)
    elif file_path.endswith((".tar", ".tar.gz", ".tgz")):
        with tarfile.open(file_path, "r") as tar_ref:
            for member in tqdm(members, desc="Extracting", unit="file"):
                if archive_contains_a_root_folder:
                    tar_ref.extract(member, target_dir)
                else:
                    tar_ref.extract(member, unpacked_dir)

    print(f"Archive unpacked to {unpacked_dir}\n")
    return unpacked_dir


def count_files(directory):
    total_files = 0
    for root, dirs, files in os.walk(directory):
        total_files += len(files)
    return total_files


def video_capture(src, frame_processing_func=None):
    # Open the video file
    cap = cv2.VideoCapture(src)

    # Get the frames per second (FPS) of the video
    fps = cap.get(cv2.CAP_PROP_FPS)

    # Calculate the delay in seconds between frames
    frame_delay = 1 / fps

    if not cap.isOpened():
        print("Error: Could not open video.")
        exit()

    while cap.isOpened():
        ret, frame = cap.read()
        if not ret:
            break

        if frame_processing_func:
            # Process the frame (e.g., convert to grayscale)
            frame = frame_processing_func(frame)

        # Display the frame
        cv2.imshow("Frame", frame)

        # Wait for the appropriate time based on the FPS
        key = cv2.waitKey(int(frame_delay * 1000))  # Convert seconds to milliseconds
        if key == ord("q"):  # Press 'q' to exit
            break

    cap.release()
    cv2.destroyAllWindows()


def create_download_link(file_path):
    """Create a download link for the given file path."""
    return FileLink(file_path)


def extract_num_from_end(filename):
    filename_wo_ext = os.path.splitext(filename)[0]
    extracted_num = ""
    for c in filename_wo_ext[::-1]:
        try:
            int(c)
            extracted_num += c
        except:
            break
    
    return int(extracted_num[::-1])


def get_datetime_str():
    # Get the current datetime
    current_time = datetime.now()
    # Format the datetime string
    datetime_str = current_time.strftime("%Y%m%d_%H%M%S")
    return datetime_str



import tensorflow as tf
from .validation import validate_array_like


# def handle_symbolic_tensor(out_dtypes=None):
#     def decorator(func):
#         def wrapper(*args, **kwargs):
#             nonlocal out_dtypes
#             out_dtypes = out_dtypes or list(map(lambda x: x.dtype, args))
#             validate_array_like(out_dtypes, "out_dtypes")
#             out_dtypes = out_dtypes[0] if len(out_dtypes)==1 else out_dtypes
#             return tf.numpy_function(func, [*args], out_dtypes)
#         return wrapper
#     return decorator


def handle_symbolic_tensor(func):
    def wrapper(X):
        processed = tf.numpy_function(func, [X], tf.float32)
        if X.ndim == 3:
            processed.set_shape((None, *X.shape))
        else:
            processed.set_shape(X.shape)
        return processed

    return wrapper



cifar10_classes = [
    "airplane",
    "automobile",
    "bird",
    "cat",
    "deer",
    "dog",
    "frog",
    "horse",
    "ship",
    "truck",
]
tf_built_in_random_transformations = [
    "RandomBrightness",
    "RandomContrast",
    "RandomCrop",
    "RandomFlip",
    "RandomHeight",
    "RandomRotation",
    "RandomTranslation",
    "RandomWidth",
    "RandomZoom",
    "random_hue",
    "random_jpeg_quality",
    "random_saturation",
]



import os
import cv2
from tqdm import tqdm
import concurrent.futures


def trim_video(
    vid_pth,
    default_output_dir="outputs",
    output_filename=None,
    trim_st_in_s=0,
    trim_et_in_s=5,
):
    os.makedirs(default_output_dir, exist_ok=True)
    if not output_filename:
        basename = os.path.basename(vid_pth)
        basename_wo_ext, ext = os.path.splitext(basename)
        vid_pth_wo_ext = f"{default_output_dir}/{basename_wo_ext}"
        save_as = f"{vid_pth_wo_ext}_trimmed_{trim_st_in_s}_{trim_et_in_s}{ext}"
    else:
        save_as = f"{default_output_dir}/{output_filename}"

    cap = cv2.VideoCapture(vid_pth)
    fourcc = cv2.VideoWriter_fourcc(*"mp4v")
    fps = int(cap.get(cv2.CAP_PROP_FPS))
    width = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
    height = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
    frame_size = (width, height)
    out = cv2.VideoWriter(save_as, fourcc, fps, frame_size)
    frames_to_process = (trim_et_in_s - trim_st_in_s) * fps
    while frames_to_process:
        ret, frame = cap.read()
        if not ret:
            break
        out.write(frame)
        frames_to_process -= 1
    cap.release()
    out.release()
    return save_as


def vid2frames(inp_vid_pth, out_frames_dir):
    os.makedirs(out_frames_dir, exist_ok=True)
    inp_vid_name = os.path.basename(inp_vid_pth)
    inp_vid_name_wo_ext, ext = os.path.splitext(inp_vid_name)
    vidcap = cv2.VideoCapture(inp_vid_pth)
    is_success, image = vidcap.read()
    frame_number = 0

    while is_success:
        save_path_and_name = (
            f"{out_frames_dir}/{inp_vid_name_wo_ext}_frame-{frame_number}.jpg"
        )
        cv2.imwrite(save_path_and_name, image)
        is_success, image = vidcap.read()
        frame_number += 1


def vids2frames(vids_dir, frames_dir, execution_mode="multi-processing"):
    vid_pth_list = []
    frames_pth_list = []
    for dirpath, dirnames, filenames in os.walk(vids_dir):
        if len(filenames):
            frames_pth = f"{frames_dir}/{dirpath.split(vids_dir)[-1]}"
            for filename in filenames:
                filepth = f"{dirpath}/{filename}"
                vid_pth_list.append(filepth)
                frames_pth_list.append(frames_pth)

    if execution_mode == "loop":
        # loop
        for inp_vid_pth, out_frames_dir in tqdm(
            list(zip(vid_pth_list, frames_pth_list))
        ):
            vid2frames(inp_vid_pth, out_frames_dir)

    elif execution_mode == "multi-threading":
        # multi threading
        with concurrent.futures.ThreadPoolExecutor() as executor:
            results = executor.map(vid2frames, vid_pth_list, frames_pth_list)
            for _ in tqdm(results, total=len(vid_pth_list)):
                pass

    elif execution_mode == "multi-processing":
        # multi processing
        with concurrent.futures.ProcessPoolExecutor() as executor:
            results = executor.map(vid2frames, vid_pth_list, frames_pth_list)
            for _ in tqdm(results, total=len(vid_pth_list)):
                pass



import tensorflow as tf


DepthNet = tf.keras.Sequential(
    [
        tf.keras.layers.Input(shape=(32, 32, 3)),
        # Block 1
        tf.keras.layers.Conv2D(32, 3, activation="relu", padding="same"),
        tf.keras.layers.BatchNormalization(),
        tf.keras.layers.DepthwiseConv2D(
            3, padding="same", activation="relu", depth_multiplier=2
        ),
        tf.keras.layers.Dropout(0.1),
        # Block 2
        tf.keras.layers.Conv2D(64, 3, activation="relu", strides=2, padding="same"),
        tf.keras.layers.BatchNormalization(),
        tf.keras.layers.Dropout(0.1),
        # Block 3
        tf.keras.layers.Conv2D(128, 3, activation="relu", padding="same"),
        tf.keras.layers.BatchNormalization(),
        tf.keras.layers.DepthwiseConv2D(
            3, padding="same", activation="relu", depth_multiplier=2
        ),
        tf.keras.layers.Dropout(0.3),
        # Block 4
        tf.keras.layers.Conv2D(256, 3, activation="relu", strides=2, padding="same"),
        tf.keras.layers.BatchNormalization(),
        tf.keras.layers.DepthwiseConv2D(3, padding="same", activation="relu"),
        tf.keras.layers.Dropout(0.3),
        # Block 5
        tf.keras.layers.Conv2D(512, 3, activation="relu", strides=2, padding="same"),
        tf.keras.layers.BatchNormalization(),
        tf.keras.layers.Dropout(0.4),
        # Dense Layers
        tf.keras.layers.Flatten(),
        tf.keras.layers.Dense(2048, activation="relu"),
        tf.keras.layers.Dropout(0.4),
        tf.keras.layers.Dense(512, activation="relu"),
        tf.keras.layers.Dropout(0.3),
        tf.keras.layers.Dense(10, activation="softmax"),
    ]
)


def all_cnn_model(input_shape=(32, 32, 3), num_classes=10):
    model = tf.keras.Sequential()

    # First convolution block
    model.add(
        tf.keras.layers.Conv2D(
            96,
            kernel_size=3,
            padding="same",
            activation="relu",
            input_shape=input_shape,
        )
    )
    model.add(
        tf.keras.layers.Conv2D(96, kernel_size=3, padding="same", activation="relu")
    )
    model.add(tf.keras.layers.MaxPooling2D(pool_size=2))

    # Second convolution block
    model.add(
        tf.keras.layers.Conv2D(192, kernel_size=3, padding="same", activation="relu")
    )
    model.add(
        tf.keras.layers.Conv2D(192, kernel_size=3, padding="same", activation="relu")
    )
    model.add(tf.keras.layers.MaxPooling2D(pool_size=2))

    # Third convolution block
    model.add(
        tf.keras.layers.Conv2D(192, kernel_size=3, padding="same", activation="relu")
    )
    model.add(
        tf.keras.layers.Conv2D(192, kernel_size=3, padding="same", activation="relu")
    )
    model.add(tf.keras.layers.MaxPooling2D(pool_size=2))

    # Fully Convolutional Layers (without fully connected layers)
    model.add(tf.keras.layers.Conv2D(10, kernel_size=1))

    # Flatten and apply softmax
    model.add(tf.keras.layers.GlobalAveragePooling2D())
    model.add(tf.keras.layers.Activation("softmax"))

    return model



import os
import cv2
from mlu_tools.utils import extract_num_from_end, get_dynamic_path
import shutil
import random
import numpy as np
import tensorflow as tf
import pandas as pd
import math


def frames2vid(frames_dir, output_video_path, fps):
    frame = cv2.imread(f"{frames_dir}/{os.listdir(frames_dir)[0]}")
    fourcc = cv2.VideoWriter_fourcc(*"mp4v")  # Codec for .mp4
    frame_size = frame.shape[1::-1]
    out = cv2.VideoWriter(output_video_path, fourcc, fps, frame_size)
    for filename in sorted(os.listdir(frames_dir), key=extract_num_from_end):
        file_path = f"{frames_dir}/{filename}"
        frame = cv2.imread(file_path)
        out.write(frame)
    out.release()


def perform_undersampling(dir_pth):
    shutil.copytree(dir_pth, f"{dir_pth}_undersampled")
    dir_pth = f"{dir_pth}_undersampled"
    min_files_cnt = math.inf
    for subdir in os.listdir(dir_pth):
        subdir_pth = f"{dir_pth}/{subdir}"
        min_files_cnt = min(min_files_cnt, len(os.listdir(subdir_pth)))

    for subdir in os.listdir(dir_pth):
        subdir_pth = f"{dir_pth}/{subdir}"
        filenames = os.listdir(subdir_pth)
        remove_filenames = random.sample(
            os.listdir(subdir_pth), len(filenames) - min_files_cnt
        )
        for filename in remove_filenames:
            filepath = f"{subdir_pth}/{filename}"
            os.remove(filepath)


def perform_oversampling(dir_pth, target_size):
    if not os.path.exists(f"{dir_pth}_oversampled"):
        shutil.copytree(dir_pth, f"{dir_pth}_oversampled")
    dir_pth = f"{dir_pth}_oversampled"
    max_files_cnt = -math.inf
    for subdir in os.listdir(dir_pth):
        subdir_pth = f"{dir_pth}/{subdir}"
        max_files_cnt = max(max_files_cnt, len(os.listdir(subdir_pth)))

    datagen = tf.keras.preprocessing.image.ImageDataGenerator(
        rotation_range=20,  # Randomly rotate images by up to 20 degrees
        width_shift_range=0.2,  # Randomly shift images horizontally by up to 20% of width
        height_shift_range=0.2,  # Randomly shift images vertically by up to 20% of height
        shear_range=20,  # Apply shear transformations with a shear angle of 20 degrees
        zoom_range=0.2,  # Randomly zoom images in/out by up to 20%
        horizontal_flip=True,  # Randomly flip images horizontally
    )

    for subdir in os.listdir(dir_pth):
        subdir_pth = f"{dir_pth}/{subdir}"
        filenames = os.listdir(subdir_pth)
        file_paths = [os.path.join(subdir_pth, img) for img in filenames]
        df = pd.DataFrame({"filename": file_paths})
        num_img_in_dir = len(filenames)
        num_img_to_gen = max_files_cnt - num_img_in_dir
        image_generator = datagen.flow_from_dataframe(
            dataframe=df,
            x_col="filename",
            y_col=None,
            class_mode=None,
            batch_size=num_img_to_gen,
            target_size=target_size,
        )
        num_iters = int(np.ceil(num_img_to_gen / num_img_in_dir))
        while num_iters:
            images = next(image_generator)

            # for images in image_generator: break
            for image in images:
                filename = "oversampled.jpg"
                filepath = f"{subdir_pth}/{filename}"
                cv2.imwrite(get_dynamic_path(filepath), image[..., ::-1])

            num_iters -= 1


def process_path(file_path, image_size, crop_to_aspect_ratio):
    label = tf.strings.split(file_path, os.sep)[-2]
    # label = int(label)
    label = tf.strings.to_number(label, out_type=tf.int32)
    img = tf.io.read_file(file_path)
    img = tf.image.decode_jpeg(img, channels=3)
    # img = tf.image.resize(img, image_size)  # doesn't support center cropping
    img = tf.keras.layers.Resizing(
        *image_size, crop_to_aspect_ratio=crop_to_aspect_ratio
    )(img)

    return img, label


def get_affine_transform_vector(
    theta, tx, ty, shear, zx, zy, image_height, image_width
):
    # theta = math.radians(theta)
    theta = tf.math.multiply(theta, tf.constant(math.pi / 180.0))
    # shear = math.radians(shear)
    shear = tf.math.multiply(shear, tf.constant(math.pi / 180.0))

    # Center of the image
    cx, cy = image_width / 2.0, image_height / 2.0

    # Rotation
    rotation = tf.convert_to_tensor(
        [[tf.cos(theta), -tf.sin(theta)], [tf.sin(theta), tf.cos(theta)]],
        dtype=tf.float32,
    )

    # Shear
    shear_matrix = tf.convert_to_tensor(
        [[1.0, -tf.sin(shear)], [0.0, tf.cos(shear)]], dtype=tf.float32
    )

    # Zoom
    zoom = tf.convert_to_tensor([[zx, 0.0], [0.0, zy]], dtype=tf.float32)

    # Composite transform
    transform = tf.linalg.matmul(rotation, tf.linalg.matmul(shear_matrix, zoom))

    # Offset for center + translation
    offset = tf.convert_to_tensor(
        [
            cx - (transform[0, 0] * cx + transform[0, 1] * cy) + tx,
            cy - (transform[1, 0] * cx + transform[1, 1] * cy) + ty,
        ],
        dtype=tf.float32,
    )

    # 8-element vector with 2 zeros at the end
    flat_transform = tf.stack(
        [
            transform[0, 0],
            transform[0, 1],
            offset[0],
            transform[1, 0],
            transform[1, 1],
            offset[1],
            0.0,
            0.0,
        ]
    )

    return flat_transform


def apply_affine_transform_tf(
    x,
    rotation_range=0,
    height_shift_range=0,
    width_shift_range=0,
    shear_range=0,
    zoom_range=0,
    horizontal_flip=False,
    fill_mode="NEAREST",
    interpolation="BILINEAR",
    fill_value=0.0,
):
    """
    Apply affine transform to a single image tensor (H, W, C).
    """

    # print(x.shape)

    if rotation_range:
        # theta = np.random.uniform(-rotation_range, rotation_range)
        theta = tf.random.uniform([], minval=-rotation_range, maxval=rotation_range)
    else:
        theta = 0.0

    if width_shift_range:
        ty = tf.random.uniform([], -width_shift_range, width_shift_range)
        ty *= tf.cast(tf.shape(x)[1], tf.float32)
    else:
        ty = tf.constant(0.0, dtype=tf.float32)

    if height_shift_range:
        tx = tf.random.uniform([], -height_shift_range, height_shift_range)
        tx *= tf.cast(tf.shape(x)[0], tf.float32)
    else:
        tx = tf.constant(0.0, dtype=tf.float32)

    if shear_range:
        shear = tf.random.uniform([], -shear_range, shear_range)
    else:
        shear = 0.0

    zoom_range = [1 - zoom_range, 1 + zoom_range]
    if zoom_range[0] == 1 and zoom_range[1] == 1:
        zx, zy = 1, 1
    else:
        ## Non-Uniform Zooming across Width and Height
        zx, zy = tf.unstack(tf.random.uniform([2], zoom_range[0], zoom_range[1]))
        ## Uniform Zooming across Width and Height
        # zx = tf.random.uniform([], zoom_range[0], zoom_range[1])
        # zy = zx

    if horizontal_flip and tf.random.uniform([], 0.0, 1.0) < 0.5:
        x = tf.reverse(x, axis=[1])

    if len(x.shape) != 3:
        raise ValueError("Input must be a 3D tensor [height, width, channels]")

    shape = tf.shape(x)
    height, width = shape[0], shape[1]

    transform = get_affine_transform_vector(
        theta,
        tx,
        ty,
        shear,
        zx,
        zy,
        tf.cast(height, tf.float32),
        tf.cast(width, tf.float32),
    )

    transform = tf.reshape(transform, [1, 8])  # shape: [1, 8]
    x = tf.expand_dims(x, axis=0)  # shape: [1, H, W, C]

    transformed = tf.raw_ops.ImageProjectiveTransformV3(
        images=x,
        transforms=transform,
        output_shape=tf.stack([height, width]),
        interpolation=interpolation,
        fill_mode=fill_mode,
        fill_value=fill_value,
    )

    return tf.squeeze(transformed, axis=0)  # shape: [H, W, C]


class TFDataGenerator:
    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def flow_from_directory(
        self,
        dir_pth,
        shuffle=True,
        image_size=(256, 256),
        crop_to_aspect_ratio=False,
        batch_size=32,
    ):
        dataset = tf.data.Dataset.list_files(f"{dir_pth}/*/*", shuffle=shuffle)
        dataset = dataset.map(
            lambda x: process_path(x, image_size, crop_to_aspect_ratio),
            num_parallel_calls=tf.data.AUTOTUNE,
        )
        if self.kwargs:
            dataset = dataset.map(
                lambda x, y: (apply_affine_transform_tf(x, **self.kwargs), y),
                num_parallel_calls=tf.data.AUTOTUNE,
            ).repeat()
        dataset = dataset.batch(
            batch_size, num_parallel_calls=tf.data.AUTOTUNE
        ).prefetch(buffer_size=tf.data.AUTOTUNE)
        return dataset


def random_rotate(image, rotation_range):
    """Rotate the image within the specified range and fill blank space with nearest neighbor."""
    # Generate a random rotation angle in radians
    theta = tf.random.uniform([], -rotation_range, rotation_range) * tf.constant(
        3.14159265 / 180, dtype=tf.float32
    )

    # Get the image dimensions
    image_shape = tf.shape(image)
    height, width = image_shape[0], image_shape[1]

    # Create the rotation matrix
    rotation_matrix = tf.stack(
        [
            [tf.cos(theta), -tf.sin(theta), 0],
            [tf.sin(theta), tf.cos(theta), 0],
            [0, 0, 1],
        ]
    )

    # Adjust for center-based rotation
    translation_to_origin = tf.stack(
        [[1, 0, -width / 2], [0, 1, -height / 2], [0, 0, 1]]
    )

    translation_back = tf.stack([[1, 0, width / 2], [0, 1, height / 2], [0, 0, 1]])

    # Cast matrices to tf.float32 for compatibility
    rotation_matrix = tf.cast(rotation_matrix, tf.float32)
    translation_to_origin = tf.cast(translation_to_origin, tf.float32)
    translation_back = tf.cast(translation_back, tf.float32)

    # Perform matrix multiplication
    transform_matrix = tf.linalg.matmul(
        translation_back, tf.linalg.matmul(rotation_matrix, translation_to_origin)
    )

    # Extract the affine part of the transformation matrix (2x3 matrix for 2D transformation)
    affine_matrix = transform_matrix[:2, :]

    # Flatten the matrix into a 1D array and add [0, 0] to make it 8 elements
    affine_matrix_8 = tf.concat(
        [affine_matrix[0, :], affine_matrix[1, :], [0, 0]], axis=0
    )

    # Apply the transformation with `fill_mode="nearest"`
    rotated_image = tf.raw_ops.ImageProjectiveTransformV2(
        images=tf.expand_dims(image, 0),
        transforms=tf.reshape(affine_matrix_8, [1, 8]),
        output_shape=tf.shape(image)[:2],
        interpolation="BILINEAR",
        fill_mode="NEAREST",
    )

    return tf.squeeze(rotated_image)


def random_translate(image, width_factor, height_factor):
    """Randomly translate the image horizontally and vertically within the specified factors.

    Args:
        image: Input image tensor.
        width_factor: Horizontal shift factor (0.1 means 10% of width).
        height_factor: Vertical shift factor (0.1 means 10% of height).

    Returns:
        Translated image.
    """
    # Get the image dimensions
    image_shape = tf.shape(image)
    height, width = image_shape[0], image_shape[1]

    # Convert factors to tensors and cast them to float32
    width_factor = tf.cast(width_factor, tf.float32)
    height_factor = tf.cast(height_factor, tf.float32)

    # Cast image dimensions to float32 to match the factor types
    width = tf.cast(width, tf.float32)
    height = tf.cast(height, tf.float32)

    # Calculate the maximum shifts based on the image dimensions
    max_width_shift = width * width_factor
    max_height_shift = height * height_factor

    # Generate random translation values within the given factors
    tx = tf.random.uniform([], -max_width_shift, max_width_shift, dtype=tf.float32)
    ty = tf.random.uniform([], -max_height_shift, max_height_shift, dtype=tf.float32)

    # Create the translation matrix as a 1D array with 8 values
    # [a, b, tx, d, e, ty, 0, 0]
    translation_matrix = tf.concat(
        [
            tf.ones([1], dtype=tf.float32),  # a = 1
            tf.zeros([1], dtype=tf.float32),  # b = 0
            [tx],  # tx (horizontal shift)
            tf.zeros([1], dtype=tf.float32),  # d = 0
            tf.ones([1], dtype=tf.float32),  # e = 1
            [ty],  # ty (vertical shift)
            tf.zeros([2], dtype=tf.float32),  # [0, 0]
        ],
        axis=0,
    )

    # Apply the translation with `fill_mode="nearest"`
    translated_image = tf.raw_ops.ImageProjectiveTransformV2(
        images=tf.expand_dims(image, 0),
        transforms=tf.reshape(translation_matrix, [1, 8]),  # Ensure 8 values
        output_shape=tf.shape(image)[:2],
        interpolation="BILINEAR",
        fill_mode="NEAREST",
    )

    return tf.squeeze(translated_image)



class TreeNode:
    def __init__(self, value):
        self.value = value  # Node's value
        self.children = []  # List to store child nodes

    def add_child(self, child):
        self.children.append(child)  # Add a child node

    def display(self, prefix="", is_last=True):
        # Print the current node with pipes and branches
        connector = "└── " if is_last else "├── "
        print(prefix + connector + self.value)

        # Update the prefix for children
        new_prefix = prefix + ("    " if is_last else "│   ")

        # Recursively display children
        for i, child in enumerate(self.children):
            is_last_child = i == len(self.children) - 1
            child.display(new_prefix, is_last_child)



from setuptools import setup, find_packages
import codecs
import os
import re

here = os.path.abspath(os.path.dirname(__file__))

with codecs.open(os.path.join(here, "README.md"), encoding="utf-8") as fh:
    long_description = "\n" + fh.read()

VERSION = "0.0.39"
DESCRIPTION = "Contains useful functions and classes"

HYPHEN_E_DOT = r"-e .+"


def get_requirements(file_path):
    with open(file_path) as file_obj:
        requirements = file_obj.readlines()
        requirements = [
            req.replace("\n", "")
            for req in requirements
            if not re.search(HYPHEN_E_DOT, req)
        ]
    return requirements


# Setting up
setup(
    name="mlu_tools",
    version=VERSION,
    author="Vikas Sanwal",
    author_email="<vikassnwl@gmail.com>",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=long_description,
    packages=find_packages(),
    install_requires=get_requirements("requirements.txt"),
    keywords=["python"],
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ],
)



import os
import tensorflow as tf


class CustomModelCheckpoint(tf.keras.callbacks.Callback):
    def __init__(self, filepath, monitor="val_accuracy", mode="max"):
        super().__init__()
        self.filepath = filepath
        self.monitor = monitor
        self.mode = mode
        self.best = None
        self.latest_checkpoint = None

        # Ensure the directory exists (if any directory is part of the path)
        directory = os.path.dirname(filepath)
        if directory:  # Only create if directory is specified
            os.makedirs(directory, exist_ok=True)

    def on_epoch_end(self, epoch, logs=None):
        if logs is None:
            logs = {}

        # Get the monitored metric
        current = logs.get(self.monitor)
        if current is None:
            return

        # Determine whether the metric improved
        if (
            self.best is None
            or (self.mode == "max" and current > self.best)
            or (self.mode == "min" and current < self.best)
        ):
            self.best = current

            # Format the filepath with epoch and metric values
            filename = self.filepath.format(epoch=epoch + 1, **logs)

            # Save the new model
            self.model.save(filename)

            # Delete the previous checkpoint after saving the new one
            if self.latest_checkpoint and os.path.exists(self.latest_checkpoint):
                os.remove(self.latest_checkpoint)

            # Update the latest checkpoint
            self.latest_checkpoint = filename



import random
import numpy as np
import matplotlib.pyplot as plt
import os
from pathlib import Path
import cv2
import tensorflow as tf
from mlu_tools.validation import validate_array_like


def from_arr(
    X,
    y=None,
    y_preds=None,
    class_names=None,
    scaling_factor=2.5,
    total_items_to_show=25,
    seed=None,
    shuffle=True,
    custom_titles=None,
    title_template="{cname}",
):
    random.seed(seed)
    if y_preds is not None:
        pred_labels = y_preds.argmax(axis=-1)
        pred_confs = y_preds[np.arange(len(pred_labels)), pred_labels]
    total_items_to_show = min(len(X), total_items_to_show)
    if shuffle:
        rand_indices = random.sample(range(len(X)), total_items_to_show)
    else:
        rand_indices = range(total_items_to_show)
    cols = int(np.ceil(np.sqrt(total_items_to_show)))
    rows = int(np.ceil(total_items_to_show / cols))
    fig, axes = plt.subplots(
        rows, cols, figsize=(scaling_factor * cols, scaling_factor * rows)
    )
    img_h, img_w = X[0].shape[:2]
    true_class_y = max(img_h, img_w) * 3 / 32
    for i, rand_idx in enumerate(rand_indices):
        ax = axes[i // cols][i % cols]
        image = X[rand_idx]
        ax.imshow(image)
        if custom_titles is not None:
            ax.set_title(custom_titles[rand_idx])
        elif y is not None:
            if y_preds is not None:
                cname = (
                    class_names[pred_labels[rand_idx].item()]
                    if class_names
                    else pred_labels[rand_idx].item()
                )
                conf = pred_confs[rand_idx].item()
                title = title_template.format(cname=cname, conf=conf)
                if pred_labels[rand_idx].item() == y[rand_idx].item():
                    ax.set_title(title, color="green")
                else:
                    ax.set_title(title, color="red")
                    if class_names:
                        ax.text(
                            0,
                            true_class_y,
                            class_names[y[rand_idx].item()],
                            color="white",
                            bbox=dict(facecolor="green"),
                        )
                    else:
                        ax.text(
                            0,
                            true_class_y,
                            y[rand_idx].item(),
                            color="white",
                            bbox=dict(facecolor="green"),
                        )
            elif class_names is not None:
                ax.set_title(class_names[y[rand_idx].item()])
            else:
                ax.set_title(y[rand_idx].item())
        ax.axis("off")


def from_dir(
    X,
    y=None,
    y_preds=None,
    class_names=None,
    scaling_factor=2.5,
    total_items_to_show=25,
    seed=None,
    shuffle=True,
    custom_titles=None,
    title_template="{cname}",
):
    random.seed(seed)
    # Prioritize custom_titles over y and y_preds
    if custom_titles is not None or y is None:
        y = None
        y_preds = None

    infer_from_dir = "infer_from_dir"
    if isinstance(y, str):
        if y != infer_from_dir:
            raise Exception(
                f"'{y}' is an invalid value for y, did you mean '{infer_from_dir}'?"
            )
    elif validate_array_like(y, raise_exception=False):
        y = np.array(y)
        if y.ndim == 2 and y.shape[1] > 1:
            y = y.argmax(axis=1)

    directory_path = X

    # directory_items = os.listdir(directory_path)
    directory_items = []
    if y is infer_from_dir:
        y_temp = []
        label = 0
    for root, dirs, files in os.walk(directory_path):
        # directory_items.extend(files)
        if files:
            directory_items.extend(Path(root).iterdir())
            if y is infer_from_dir:
                y_temp.extend([label] * len(files))
                label += 1

    if y is infer_from_dir:
        y = np.array(y_temp)

    total_items_to_show = min(len(directory_items), total_items_to_show)
    if shuffle:
        rand_indices = random.sample(range(len(directory_items)), total_items_to_show)
    else:
        rand_indices = range(total_items_to_show)

    X = []
    if y is not None:
        y_temp = []
    if y_preds is not None:
        y_preds_temp = []
    # for item_name in directory_items[:total_items_to_show]:
    for rand_idx in rand_indices:
        # item_path = f"{directory_path}/{item_name}"
        item_path = directory_items[rand_idx]
        image = cv2.imread(item_path)[..., ::-1]
        X.append(image)
        if y is not None:
            y_temp.append(y[rand_idx])
        if y_preds is not None:
            y_preds_temp.append(y_preds[rand_idx])
    if y is not None:
        y = y_temp
    if y_preds is not None:
        y_preds = y_preds_temp

    from_arr(
        X,
        y,
        y_preds,
        class_names,
        scaling_factor,
        total_items_to_show,
        seed,
        shuffle=False,
        custom_titles=custom_titles,
        title_template=title_template,
    )


def from_ds(
    X,
    y=None,
    y_preds=None,
    class_names=None,
    scaling_factor=2.5,
    total_items_to_show=25,
    custom_titles=None,
    title_template="{cname}",
):
    i = 0
    image_ds = X
    X = []
    y = []
    for images, labels in image_ds:
        if i == total_items_to_show:
            break
        items_to_fetch_from_batch = min(total_items_to_show - i, len(images))
        X.extend(tf.cast(images[:items_to_fetch_from_batch], "uint8"))
        y.extend(labels[:items_to_fetch_from_batch].numpy())
        i += items_to_fetch_from_batch

    from_arr(
        X,
        y,
        y_preds,
        class_names,
        scaling_factor,
        total_items_to_show,
        shuffle=False,
        custom_titles=custom_titles,
        title_template=title_template,
    )


def from_paths(
    X,
    y=None,
    y_preds=None,
    class_names=None,
    scaling_factor=2.5,
    total_items_to_show=25,
    seed=None,
    shuffle=True,
    custom_titles=None,
    title_template="{cname}",
):
    paths = X
    X = []
    for path in paths:
        X.append(cv2.imread(path)[..., ::-1])

    from_arr(
        X,
        y,
        y_preds,
        class_names,
        scaling_factor,
        total_items_to_show,
        seed,
        shuffle=False,
        custom_titles=custom_titles,
        title_template=title_template,
    )



import os
from collections.abc import Iterable
import inspect
from pathlib import Path


def validate_file(path, valid_types=[], raise_exception=True):
    param_name = "path"
    if not os.path.exists(path):
        if raise_exception:
            raise Exception(f"{param_name}='{path}' does not exist!")
        return False

    if not os.path.isfile(path):
        if raise_exception:
            raise Exception(f"{param_name}='{path}' is not a file!")
        return False

    if not valid_types:
        return True

    for valid_type in valid_types:
        path_wo_ext, ext = os.path.splitext(path)
        ext = ext.lower()
        if valid_type == "vid":
            video_extensions = {
                ".mp4",
                ".avi",
                ".mkv",
                ".mov",
                ".wmv",
                ".flv",
                ".webm",
                ".mpeg",
            }
            if ext in video_extensions:
                break
        elif valid_type == "img":
            image_extensions = {
                ".jpg",
                ".jpeg",
                ".png",
                ".bmp",
                ".tiff",
                ".gif",
                ".webp",
            }
            if ext in image_extensions:
                break
        elif valid_type == "txt":
            if ext == ".txt":
                break
        elif valid_type == "csv":
            if ext == ".csv":
                break
        else:
            raise Exception(
                f"type='{valid_type}' is not implemented for function `{validate_file.__name__}`"
            )
    else:
        if raise_exception:
            raise Exception(
                f"{param_name}='{path}' is not in the specified valid_types:\n{valid_types}"
            )
        return False

    return valid_type


def validate_dir(path):
    param_name = "path"
    if not isinstance(path, str):
        raise Exception(f"'{param_name}' must be a string not {type(path).__name__}.")
    if not os.path.exists(path):
        raise Exception(f"{param_name}='{path}' does not exist!")
    if not os.path.isdir(path):
        raise Exception(f"{param_name}='{path}' is not a directory!")


def validate_array_like(obj, raise_exception=True, custom_message=""):
    param_name = "obj"
    custom_message = (
        custom_message
        or f"'{param_name}' must be an array-like object i.e., list, tuple, np.ndarray, etc., not {type(obj).__name__}."
    )
    if not isinstance(obj, Iterable) or isinstance(obj, str):
        if raise_exception:
            raise Exception(custom_message)
        return False
    return True


def validate_pixel_range_0_255(X, custom_message=""):
    param_name = "X"
    if X.min() < 0 or X.max() <= 1:
        default_message = (
            f"Pixel values are not in the range 0..255 for argument '{param_name}'."
        )
        message = custom_message or default_message
        raise Exception(message)


def validate_dir_structure(dir_path, structure="dir>subdir>img"):
    validate_dir(dir_path)
    if structure == "dir>subdir>img":
        for i, (root, dirs, files) in enumerate(os.walk(dir_path)):
            if i == 0:
                if not dirs or files:
                    raise Exception(
                        f"Structure didn't match the specified one: {structure}"
                    )
            else:
                if (
                    dirs
                    or not files
                    or not [validate_file(x, ["img"]) for x in Path(root).iterdir()]
                ):
                    raise Exception(
                        f"Structure didn't match the specified one: {structure}"
                    )
